#
# Copyright 2021, Flyreel. All Rights Reserved.
# =============================================#
"""Loss functions and modules."""
import torch.nn.functional as F

from torch import Tensor


def softmax_mse_loss(inputs: Tensor, targets: Tensor) -> Tensor:
    """Computes the mean squared error loss after softmax."""

    return F.mse_loss(F.softmax(inputs, dim=-1), F.softmax(targets, dim=-1), reduction="none")


def kl_div_loss(inputs: Tensor, targets: Tensor) -> Tensor:
    """Computes the Kullback–Leibler divergence loss between two probability distributions."""

    return F.kl_div(F.log_softmax(inputs, dim=-1), F.softmax(targets, dim=-1), reduction="none")




#
# Copyright 2021, Flyreel. All Rights Reserved.
# =============================================#
"""SESEMI learners."""
from typing import Optional, Tuple

from torch import Tensor
import torch
import torch.nn as nn
import torch.nn.functional as F

import copy
import logging
import os.path as osp
import pytorch_lightning as pl

from pytorch_lightning.trainer.states import RunningStage

from hydra.utils import instantiate
from torchmetrics.classification.accuracy import Accuracy
from torchmetrics import MeanMetric

from .config.structs import ClassifierHParams
from .models.backbones.base import Backbone
from .models.heads.base import LinearHead
from .utils import reduce_tensor
from .schedulers.weight import WeightScheduler

logger = logging.getLogger(__name__)


class Classifier(pl.LightningModule):
    """The SESEMI classifier."""

    hparams: ClassifierHParams  # type: ignore

    def __init__(self, hparams: ClassifierHParams):
        """Initializes the module.

        Args:
            hparams: The classifier's hyperparameters.
        """
        super().__init__()
        self.save_hyperparameters(hparams)

        # Instantiate shared nn.Modules
        self.shared_backbones = nn.ModuleDict()
        self.shared_backbones["backbone"] = instantiate(hparams.model.backbone)

        self.shared_heads = nn.ModuleDict()
        self.shared_heads["supervised"] = LinearHead(
            self.backbone.out_features, hparams.num_classes
        )

        # Instantiate the supervised loss, and its scheduler and reduction options.
        self.supervised_loss = instantiate(
            hparams.model.supervised_loss.callable, reduction="none"
        )
        if hparams.model.supervised_loss.scheduler is not None:
            self.supervised_loss_scheduler: Optional[WeightScheduler] = instantiate(
                hparams.model.supervised_loss.scheduler
            )
        else:
            self.supervised_loss_scheduler = None
        self.supervised_loss_reduction_method = hparams.model.supervised_loss.reduction
        self.ema = hparams.model.ema

        self.register_buffer(
            "best_validation_top1_accuracy",
            torch.tensor(0.0, dtype=torch.float32, device=self.device),
        )
        self.register_buffer(
            "best_ema_validation_top1_accuracy",
            torch.tensor(0.0, dtype=torch.float32, device=self.device),
        )

        # Instantiate the regularization loss heads,
        # and their optional scheduler and reduction methods.
        self.regularization_loss_heads = nn.ModuleDict()
        self.regularization_loss_schedulers = {}
        self.regularization_loss_reduction_methods = {}

        regularization_loss_head_configs = hparams.model.regularization_loss_heads or {}
        for name, loss_head_config in regularization_loss_head_configs.items():
            self.regularization_loss_heads[name] = instantiate(
                loss_head_config.head, logger=self.logger
            )
            if loss_head_config.scheduler is not None:
                self.regularization_loss_schedulers[name] = instantiate(
                    loss_head_config.scheduler
                )
            else:
                self.regularization_loss_schedulers[name] = None
            self.regularization_loss_reduction_methods[name] = (
                loss_head_config.reduction or "mean"
            )
        self.num_regularization_losses = len(self.regularization_loss_heads)

        if self.ema is not None:
            assert (
                0.0 <= self.ema.decay <= 1.0
            ), "EMA decay value should be between [0, 1]. Default 0.999."
            self.shared_backbones["backbone_ema"] = self._copy_and_detach(self.backbone)
            self.shared_heads["supervised_ema"] = self._copy_and_detach(self.head)

        # Build the regularization loss heads.
        for head in self.regularization_loss_heads.values():
            head.build(self.shared_backbones, self.shared_heads)

        # Initialize training and validation metrics.
        self.training_accuracy = Accuracy(top_k=1, dist_sync_on_step=True)
        self.validation_top1_accuracy = Accuracy(top_k=1)
        self.validation_average_loss = MeanMetric()
        self.ema_validation_top1_accuracy = Accuracy(top_k=1)
        self.ema_validation_average_loss = MeanMetric()

    def _copy_and_detach(self, module):
        """Detaches a new module from the computational graph after copying."""
        new_module = copy.deepcopy(module)
        for param in new_module.parameters():
            param.detach_()
        return new_module

    @property
    def backbone(self) -> Backbone:
        """The supervised backbone."""
        return self.shared_backbones["backbone"]

    @property
    def backbone_ema(self) -> Backbone:
        """The supervised backbone with EMA weights."""
        return self.shared_backbones["backbone_ema"]

    @property
    def head(self) -> LinearHead:
        """The supervised head."""
        return self.shared_heads["supervised"]

    @property
    def head_ema(self) -> LinearHead:
        """The supervised head with EMA weights."""
        return self.shared_heads["supervised_ema"]

    def forward(self, x):
        features = self.backbone(x)
        logits = self.head(features)
        return logits

    def configure_optimizers(self):
        optimizer = instantiate(self.hparams.optimizer, params=self.parameters())

        if self.hparams.lr_scheduler is not None:
            lr_dict = dict(self.hparams.lr_scheduler)
            lr_dict["scheduler"] = instantiate(
                lr_dict["scheduler"], optimizer=optimizer
            )
            return [optimizer], [lr_dict]

        return optimizer

    def training_step(self, batch, batch_index):
        shared_features = {}

        inputs_t, targets_t = batch["supervised"][:2]
        features_t = self.backbone(inputs_t)
        outputs_t = self.head(features_t)

        shared_features["backbone"] = features_t
        shared_features["supervised"] = outputs_t

        supervised_loss = self.supervised_loss(outputs_t, targets_t)

        regularization_losses = {
            name: head(
                data=batch,
                backbones=self.shared_backbones,
                heads=self.shared_heads,
                features=shared_features,
                step=self.global_step,
            )
            for name, head in self.regularization_loss_heads.items()
        }

        return {
            "supervised_loss": supervised_loss,
            "regularization_losses": regularization_losses,
            "probs": F.softmax(outputs_t, dim=-1),
            "targets": targets_t,
        }

    def _compute_weighted_training_loss(
        self,
        loss: Tensor,
        reduction: Optional[str],
        scheduler: Optional[WeightScheduler],
        scale_factor: float,
        log_prefix: Optional[str] = None,
    ) -> Tuple[Tensor, Tensor, float]:
        reduced_loss = reduce_tensor(loss, reduction)

        if scheduler is not None:
            scheduler_weight = scheduler(self.global_step)
        else:
            scheduler_weight = 1.0

        loss_weight = scale_factor * scheduler_weight
        weighted_loss = loss_weight * reduced_loss

        if log_prefix is not None:
            self.log(osp.join(log_prefix, "loss"), reduced_loss, on_step=True)
            self.log(osp.join(log_prefix, "loss_weight"), loss_weight, on_step=True)
            self.log(osp.join(log_prefix, "weighted_loss"), weighted_loss, on_step=True)

        return reduced_loss, weighted_loss, loss_weight

    def _log_learning_rates(self):
        optim = self.optimizers()
        param_group0_lr = optim.optimizer.param_groups[0]["lr"]
        for i, param_group in enumerate(optim.optimizer.param_groups):
            self.log(f"optim/lr/param_group/{i}", param_group["lr"], on_step=True)

        self.log(
            "lr",
            param_group0_lr,
            on_step=True,
            on_epoch=False,
            prog_bar=True,
            logger=False,
        )

    def _ema_update(self, ema_module, module, step, decay):
        """Computes in-place the EMA parameters from the original parameters."""
        # Use the true average until the exponential average is more correct.
        decay = min(1.0 - 1.0 / (step + 1.0), decay)
        for ema_param, param in zip(ema_module.parameters(), module.parameters()):
            ema_param.data.mul_(decay).add_(param.data, alpha=(1.0 - decay))

    def training_step_end(self, outputs):
        _, weighted_supervised_loss, _ = self._compute_weighted_training_loss(
            loss=outputs["supervised_loss"],
            reduction=self.supervised_loss_reduction_method,
            scheduler=self.supervised_loss_scheduler,
            scale_factor=self.hparams.model.supervised_loss.scale_factor,
            log_prefix="train/supervised",
        )

        weighted_regularization_losses = []
        for name, regularization_loss in outputs["regularization_losses"].items():
            _, weighted_regularization_loss, _ = self._compute_weighted_training_loss(
                loss=regularization_loss,
                reduction=self.regularization_loss_reduction_methods[name],
                scheduler=self.regularization_loss_schedulers[name],
                scale_factor=self.hparams.model.regularization_loss_heads[
                    name
                ].scale_factor,
                log_prefix=f"train/regularization/{name}",
            )

            weighted_regularization_losses.append(weighted_regularization_loss)

        losses = [weighted_supervised_loss] + weighted_regularization_losses
        loss = torch.sum(torch.stack(losses))

        self.training_accuracy.update(outputs["probs"], outputs["targets"])

        self.log("train/loss", loss)

        self._log_learning_rates()

        if self.ema is not None:
            self._ema_update(
                self.backbone_ema, self.backbone, self.global_step, self.ema.decay
            )
            self._ema_update(self.head_ema, self.head, self.global_step, self.ema.decay)

        return loss

    def training_epoch_end(self, outputs) -> None:
        self.log(
            "train/top1",
            self.training_accuracy.compute(),
            on_step=False,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )

        self.training_accuracy.reset()

    def compute_validation_outputs(self, inputs, targets, backbone, head):
        outputs_t = head(backbone(inputs))
        probs_t = torch.softmax(outputs_t, dim=-1)
        loss_t = F.cross_entropy(outputs_t, targets, reduction="none")
        return probs_t, loss_t

    def validation_step(self, batch, batch_index):
        inputs_t, targets_t = batch

        regular_outputs = self.compute_validation_outputs(
            inputs_t, targets_t, self.backbone, self.head
        )

        ema_outputs = (
            self.compute_validation_outputs(
                inputs_t, targets_t, self.backbone_ema, self.head_ema
            )
            if self.ema is not None
            else None
        )

        return targets_t, regular_outputs, ema_outputs

    def validation_step_end(self, outputs):
        targets_t, regular_outputs, ema_outputs = outputs

        outputs_t, loss_t = regular_outputs
        self.validation_top1_accuracy.update(outputs_t, targets_t)
        self.validation_average_loss.update(loss_t)

        if ema_outputs is not None:
            ema_outputs_t, ema_loss_t = ema_outputs
            self.ema_validation_top1_accuracy.update(ema_outputs_t, targets_t)
            self.ema_validation_average_loss.update(ema_loss_t)

    def log_validation_metrics(self, top1, best_top1, loss, prefix: str = "val"):
        self.log(
            f"{prefix}/top1",
            top1,
            on_step=False,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )

        self.log(
            f"{prefix}/top1/best",
            best_top1,
            on_step=False,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )

        self.log(
            f"{prefix}/loss",
            loss,
            on_step=False,
            on_epoch=True,
            prog_bar=True,
            logger=True,
        )

    def validation_epoch_end(self, outputs):
        top1 = self.validation_top1_accuracy.compute()
        loss = self.validation_average_loss.compute()
        self.validation_top1_accuracy.reset()
        self.validation_average_loss.reset()

        ema_top1: Optional[float] = None
        ema_loss: Optional[float] = None
        if self.ema is not None:
            ema_top1 = self.ema_validation_top1_accuracy.compute()
            ema_loss = self.ema_validation_average_loss.compute()
            self.ema_validation_top1_accuracy.reset()
            self.ema_validation_average_loss.reset()

        if self.trainer.state.stage != RunningStage.SANITY_CHECKING:
            if top1 > self.best_validation_top1_accuracy:
                self.best_validation_top1_accuracy = torch.tensor(
                    float(top1),
                    dtype=self.best_validation_top1_accuracy.dtype,
                    device=self.best_validation_top1_accuracy.device,
                )

            self.log_validation_metrics(
                top1, self.best_validation_top1_accuracy, loss, prefix="val"
            )

            if self.ema is not None:
                if ema_top1 > self.best_ema_validation_top1_accuracy:
                    self.best_ema_validation_top1_accuracy = torch.tensor(
                        float(ema_top1),
                        dtype=self.best_ema_validation_top1_accuracy.dtype,
                        device=self.best_ema_validation_top1_accuracy.device,
                    )

                self.log_validation_metrics(
                    ema_top1,
                    self.best_ema_validation_top1_accuracy,
                    ema_loss,
                    prefix="ema/val",
                )



#
# Copyright 2021, Flyreel. All Rights Reserved.
# =============================================#
import os
import numpy as np
from tqdm import trange
import logging

import torch
import hydra
from torchvision import datasets
from hydra.core.config_store import ConfigStore

from ..config.structs import SESEMIInferenceConfig
from ..learners import Classifier
from ..utils import validate_paths
from ..transforms import center_crop_transforms, multi_crop_transforms


logger = logging.getLogger(__name__)

config_store = ConfigStore.instance()
config_store.store(
    name="inference",
    node=SESEMIInferenceConfig,
    group="ops",
    package="_global_",
)


class Predictor:
    def __init__(self, model_path, classes, config):
        self.config = config
        self.model_path = model_path
        self.classes = np.array(classes)
        self.device = torch.device(
            "cpu" if config.no_cuda or not torch.cuda.is_available() else "cuda"
        )
        self._init_model()

    def _init_model(self):
        self.model = Classifier.load_from_checkpoint(
            self.model_path, map_location=self.device
        )
        assert len(self.classes) == self.model.hparams.num_classes
        logger.info(f"=> Model checkpoint loaded from {self.model_path}")
        self.model = torch.nn.DataParallel(self.model).to(self.device)
        self.model.eval()

    def predict(self, x, ncrops, topk=1):
        with torch.no_grad():
            x = x.to(self.device)
            batch_size = x.size(0)
            w, h, c = x.shape[-1:-4:-1]
            outputs = self.model(x.view(-1, c, h, w))  # fuse batch size and ncrops
            outputs = torch.softmax(outputs, dim=-1)
            outputs = outputs.view(batch_size, ncrops, -1).mean(1)  # avg over crops
            scores, indices = torch.topk(outputs, k=topk, largest=True, sorted=True)
            scores = scores.cpu().numpy()
            indices = indices.cpu().numpy()
            labels = self.classes[indices]
            return (labels, scores)


@hydra.main(config_path="./conf", config_name=None)
def predict(config: SESEMIInferenceConfig):
    # Data loading
    validate_paths([config.data_dir])
    if config.oversample:
        ncrops = config.ncrops
        test_transformations = multi_crop_transforms(
            config.resize, config.crop_dim, ncrops, interpolation=3
        )
    else:
        ncrops = 1
        test_transformations = center_crop_transforms(
            config.resize, config.crop_dim, interpolation=3
        )
    dataset = datasets.ImageFolder(config.data_dir, test_transformations)
    dataset_loader = torch.utils.data.DataLoader(
        dataset,
        batch_size=config.batch_size,
        shuffle=False,
        num_workers=config.workers,
        pin_memory=True,
        drop_last=False,
    )

    classifier = Predictor(config.checkpoint_path, dataset.classes, config)

    # Write prediction results to file
    if os.path.exists(config.outfile):
        os.remove(config.outfile)
    with open(config.outfile, "a") as f:
        header = ",".join(["Id", "Category", "Score"])
        f.write(header + "\n")
    index = 0

    dataset_iterator = iter(dataset_loader)
    for _ in trange(
        len(dataset_loader),
        desc=f"Inferencing on {len(dataset.imgs)} files",
        position=1,
    ):
        inputs, _ = next(dataset_iterator)
        labels, scores = classifier.predict(inputs, ncrops, config.topk)
        # Write prediction results to file
        with open(config.outfile, "a") as f:
            for label, score in zip(labels, scores):
                img_path = dataset.imgs[index][0]
                img_id = os.path.splitext(os.path.basename(img_path))[0]
                label = " ".join(label)
                score = [f"{s:.6f}" for s in score]
                score = " ".join(score)
                f.write(",".join([img_id, label, score]) + "\n")
                index += 1


if __name__ == "__main__":
    predict()



#
# Copyright 2021, Flyreel. All Rights Reserved.
# =============================================#
"""Utilities to download and prepare models."""
import torch

from typing import Optional

TORCH_HUB_DOWNLOAD_MAX_RETRIES = 100


def load_torch_hub_model(repo: str, model: str, *args, **kwargs):
    """Tries to load a torch hub model and handles different exceptions that could be raised.

    Args:
        repo: The GitHub repository containing the models.
        model: The model name to download.
        max_retries: The maximum number of tries to download the model.

    Returns:
        The downloaded torch model.
    """
    error: Optional[Exception] = None
    for _ in range(TORCH_HUB_DOWNLOAD_MAX_RETRIES + 1):
        try:
            try:
                return torch.hub.load(
                    repo,
                    model,
                    *args,
                    **kwargs,
                )
            except RuntimeError:
                return torch.hub.load(
                    repo,
                    model,
                    *args,
                    **kwargs,
                    force_reload=True,
                )
        except Exception as e:
            error = e

    assert error is not None
    raise error



#
# Copyright 2021, Flyreel. All Rights Reserved.
# =============================================#
"""A pseudo-labeled dataset generation operation."""
import hydra
import torch
import torch.nn as nn
import logging
import pytorch_lightning as pl
import math
import os
import h5py
import yaml
import multiprocessing as mp

from torch.utils.data import DataLoader, Subset
from collections import defaultdict

from hydra.utils import instantiate
from hydra.core.config_store import ConfigStore
from hydra.utils import to_absolute_path
from hydra.core.hydra_config import HydraConfig
from hydra.core.utils import configure_log, setup_globals

from typing import Any, Callable, ChainMap, Dict, List, Optional, TypeVar

from tqdm import tqdm
from PIL import Image
from concurrent.futures import ProcessPoolExecutor, wait

from sesemi.utils import compute_device_names

from ..config.structs import SESEMIPseudoDatasetConfig
from ..learners import Classifier

logger = logging.getLogger(__name__)


config_store = ConfigStore.instance()
config_store.store(
    name="pseudo_dataset",
    node=SESEMIPseudoDatasetConfig,
    group="ops",
    package="_global_",
)


T = TypeVar("T")


def _identity(x: T) -> T:
    """The identity function used as a default transform."""
    return x


def default_test_time_augmentation(x: T) -> List[T]:
    """An identity test-time augmentation."""
    return [x]


def _num_digits(x: int) -> int:
    """Computes the number of digits in a number base 10."""
    return int(math.ceil(math.log10(x)))


def default_image_getter(x) -> Image.Image:
    """The default image getter for a data example.

    Assumes that examples are either an image or a tuple in which the first
    element is an image (standard case).
    """
    if isinstance(x, Image.Image):
        return x
    else:
        return x[0]


def apply_model_to_test_time_augmentations(
    model: nn.Module,
    device: str,
    data: List[List[torch.Tensor]],
    batch_compatible_tensors: bool = True,
) -> List[torch.Tensor]:
    """Applies a model to test-time augmented images.

    Args:
        model: The model to apply.
        device: The device the model is on.
        data: The list of test-time-augmented images. Each
            child list corresponds to test-time augmented images
            from an original input image.
        batch_compatible_tensors: Whether to concatenate compatibly sized
            tensors.

    Returns:
        A list of batches of logits corresponding to the input images.
    """
    if batch_compatible_tensors:
        data_index_by_shape = defaultdict(list)
        for i, items in enumerate(data):
            for j, item in enumerate(items):
                data_index_by_shape[item.shape].append((i, j))

        results = []
        for indices in data_index_by_shape.values():
            tensors = [data[i][j] for i, j in indices]
            tensor_batch = torch.stack(tensors, dim=0)
            results_batch = model(tensor_batch.to(device))
            results.extend(zip(indices, list(results_batch)))
        results.sort()

        outputs = [[] for _ in range(len(data))]
        for (i, j), result in results:
            assert len(outputs[i]) == j
            outputs[i].append(result)

        return [torch.cat(x, dim=0) for x in outputs]
    else:
        return [
            torch.cat([model(tensor[None].to(device)) for tensor in tensors], dim=0)
            for tensors in data
        ]


class TestTimeAugmentationCollator:
    def __init__(
        self,
        preprocessing_transform: Optional[Callable],
        test_time_augmentation: Optional[Callable[[Image.Image], List[Image.Image]]],
        postaugmentation_transform: Optional[Callable],
        image_getter: Optional[Callable],
    ):
        """Initializes the collator.

        Args:
            preprocessing_transform: The preprocessing transform.
            test_time_augmentation: The test-time augmentation that takes an image
                and returns a list of augmented versions of that image.
            postaugmentation_transform: A transform to apply after test-time
                augmentations and which should return a tensor.
            image_getter: The function to extract the source image from
                an example in the dataset.
        """
        self.preprocessing_transform = (
            _identity if preprocessing_transform is None else preprocessing_transform
        )
        self.test_time_augmentation = (
            default_test_time_augmentation
            if test_time_augmentation is None
            else test_time_augmentation
        )
        self.postaugmentation_transform = (
            _identity
            if postaugmentation_transform is None
            else postaugmentation_transform
        )
        self.image_getter = (
            default_image_getter if image_getter is None else image_getter
        )

    def __call__(self, data_batch: List[Any]) -> List[List[torch.Tensor]]:
        """Generates test-time augmented versions of the input images."""
        data_tensors: List[List[torch.Tensor]] = []
        images: List[Image.Image] = []
        for data in data_batch:
            item: Image.Image = self.image_getter(data)
            images.append(item)

            augmentations = self.test_time_augmentation(
                self.preprocessing_transform(item)
            )
            tensors = [self.postaugmentation_transform(x) for x in augmentations]
            data_tensors.append(tensors)
        return images, data_tensors


def task(
    config: SESEMIPseudoDatasetConfig,
    hydra_config: HydraConfig,
    task_id: int,
    num_tasks: int,
    device: str,
) -> Dict[str, Dict[str, Any]]:
    """A task to generate a subset of the pseudo-labeled dataset.

    Args:
        config: The operation's configuration.
        hydra_config: The global hydra configuration.
        task_id: The assigned task id from 0 to num_tasks-1.
        num_tasks: The number of tasks running.
        device: The device to use for inference.

    Returns:
        A dictionary of detailed metadata for each image by its assigned
        string identifier.
    """
    pl.seed_everything(config.seed)

    setup_globals()
    configure_log(hydra_config.hydra.job_logging, hydra_config.hydra.verbose)
    HydraConfig.instance().set_config(hydra_config)

    learner = Classifier.load_from_checkpoint(
        to_absolute_path(config.checkpoint_path), map_location=device
    )
    learner.eval()
    learner.to(device)

    dataset = instantiate(
        config.dataset,
        image_transform=None,
    )

    collator = TestTimeAugmentationCollator(
        instantiate(config.preprocessing_transform),
        instantiate(config.test_time_augmentation),
        instantiate(config.postaugmentation_transform),
        instantiate(config.image_getter),
    )

    size = len(dataset)
    items_per_worker = math.ceil(size / num_tasks)
    start = items_per_worker * task_id
    end = min(items_per_worker * (task_id + 1), size)
    subset = Subset(dataset, list(range(start, end)))

    dataloader = DataLoader(
        subset,
        batch_size=config.batch_size,
        shuffle=False,
        num_workers=config.num_workers,
        batch_sampler=None,
        collate_fn=collator,
    )

    output_dir = to_absolute_path(config.output_dir)
    images_dir = os.path.join(output_dir, "images")
    predictions_dir = os.path.join(output_dir, "predictions")

    num_samples = len(dataset)
    num_digits = _num_digits(num_samples)

    details: Dict[str, Dict[str, Any]] = {}
    relative_index = 0
    with torch.no_grad():
        iterable = tqdm(dataloader) if task_id == 0 else dataloader
        for i, (images, data_tensors) in enumerate(iterable):
            logits = apply_model_to_test_time_augmentations(
                learner, device, data_tensors, batch_compatible_tensors=True
            )
            probabilities = [torch.softmax(l, dim=-1) for l in logits]
            avg_logits = [torch.mean(x, dim=0) for x in logits]
            avg_probabilities = [torch.mean(x, dim=0) for x in probabilities]
            for j, (image, logits_, probs) in enumerate(
                zip(images, avg_logits, avg_probabilities)
            ):
                index = start + relative_index + j
                data_id = str(index).zfill(num_digits)

                src_image_filename = image.info.get("filename")
                dst_image_filename = os.path.join(images_dir, f"{data_id}.jpg")
                if config.symlink_images and src_image_filename is not None:
                    os.symlink(src_image_filename, dst_image_filename)
                else:
                    if config.symlink_images:
                        logger.warning(
                            f"filename metadata for image {data_id} does not exist"
                        )
                    image.save(dst_image_filename)

                with h5py.File(
                    os.path.join(predictions_dir, f"{data_id}.h5"), "w"
                ) as predictions:
                    predictions.create_dataset(
                        "logits", data=logits_.detach().cpu().numpy()
                    )
                    predictions.create_dataset(
                        "probabilities", data=probs.detach().cpu().numpy()
                    )

                details[data_id] = dict(id=data_id, filename=src_image_filename)
            relative_index += len(images)

    return details


@hydra.main(config_path="./conf", config_name=None)
def pseudo_dataset(config: SESEMIPseudoDatasetConfig):
    """The pseudo dataset generator.

    Args:
        config: The pseudo dataset generation config.
    """
    mp.set_start_method("spawn")

    output_dir = to_absolute_path(config.output_dir)
    images_dir = os.path.join(output_dir, "images")
    predictions_dir = os.path.join(output_dir, "predictions")

    os.makedirs(output_dir, exist_ok=True)
    os.makedirs(images_dir, exist_ok=False)
    os.makedirs(predictions_dir, exist_ok=False)

    devices = compute_device_names(config.gpus)

    process_pool = ProcessPoolExecutor(max_workers=len(devices))

    task_futures = []
    hydra_config = HydraConfig.instance().cfg
    for task_id, device in enumerate(devices):
        task_futures.append(
            process_pool.submit(
                task,
                config=config,
                hydra_config=hydra_config,
                task_id=task_id,
                num_tasks=len(devices),
                device=device,
            )
        )

    done, not_done = wait(task_futures)
    if not_done:
        logger.error(f"Could not run inference successfully on all devices: {not_done}")
        return

    details = dict(ChainMap(*[x.result() for x in done]))
    ids = sorted(list(details.keys()))

    metadata = dict(ids=ids, details=details)
    with open(os.path.join(output_dir, "metadata.yaml"), "w") as f:
        yaml.safe_dump(metadata, f, sort_keys=False)


if __name__ == "__main__":
    pseudo_dataset()



#
# Copyright 2021, Flyreel. All Rights Reserved.
# =============================================#
"""SESEMI datasets and registry."""
import os
import yaml
import h5py
import torch
import numpy as np

from hydra.utils import to_absolute_path
from torch.utils.data import ConcatDataset, Dataset, IterableDataset
from torchvision.datasets import ImageFolder, CIFAR10, CIFAR100, STL10
from torchvision.datasets.folder import (
    default_loader,
    has_file_allowed_extension,
    IMG_EXTENSIONS,
)

from PIL import Image
from typing import Any, Callable, Dict, List, Optional, Type, Union

DatasetBuilder = Callable[..., Union[Dataset, IterableDataset]]
ImageTransform = Callable

DATASET_REGISTRY: Dict[str, DatasetBuilder] = {}


class _ImageFolder(ImageFolder):
    """An ImageFolder dataset that adds metadata to loaded PIL images."""

    def __getitem__(self, index: int):
        sample, target = super().__getitem__(index)
        if isinstance(sample, Image.Image):
            sample.info["filename"] = self.samples[index][0]
        return sample, target


class PseudoDataset(Dataset):
    def __init__(
        self,
        root: str,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        use_probability_target: bool = False,
    ):
        """A pseudo-labeled dataset.

        Args:
            root: The path to the dataset root.
            transform: An optional image transform.
            target_transform: An optional target transform.
            use_probability_target: Whether to use the probabilities
                as learning targets rather than integer labels.
        """
        super().__init__()
        self.root = root
        self.transform = transform
        self.target_transform = target_transform
        self.use_probability_target = use_probability_target

        with open(os.path.join(self.root, "metadata.yaml"), "r") as f:
            self.metadata = yaml.safe_load(f)

    def __len__(self) -> int:
        return len(self.metadata["ids"])

    def __getitem__(self, index: int):
        id_ = self.metadata["ids"][index]
        image_path = os.path.join(self.root, "images", f"{id_}.jpg")
        image = Image.open(image_path)
        image.info["filename"] = image_path

        if self.transform is not None:
            image = self.transform(image)

        prediction = h5py.File(os.path.join(self.root, "predictions", f"{id_}.h5"), "r")
        probabilities = np.array(prediction["probabilities"])
        label = probabilities.argmax()

        if self.use_probability_target:
            target = torch.tensor(probabilities)
        else:
            target = label

        return image, target


def get_image_files(directory: str, is_valid_file: Callable[[str], bool]) -> List[str]:
    """Finds the full list of image files recursively under a directory path.

    Args:
        directory: The root directory to search for image files.
        is_valid_file: A callable to determine if a file is a valid image.

    Returns:
        The list of paths to the image files.
    """
    directory = os.path.expanduser(directory)

    files: List[str] = []
    for root, _, fnames in sorted(os.walk(directory, followlinks=True)):
        for fname in sorted(fnames):
            if is_valid_file(fname):
                path = os.path.join(root, fname)
                files.append(path)

    return files


def default_is_vaild_file(path: str) -> bool:
    """The default callable to determine if a file is a valid image."""
    return has_file_allowed_extension(path, IMG_EXTENSIONS)


class ImageFile(Dataset):
    def __init__(
        self,
        root: str,
        transform: Optional[Callable] = None,
        loader: Callable[[str], Any] = default_loader,
        is_valid_file: Callable[[str], bool] = default_is_vaild_file,
    ):
        """An image-file dataset.

        Args:
            root: The path to the dataset root.
            transform: An optional image transform.
            loader: The image loading callable.
            is_valid_file: A callable to determine if a file is a valid image.
        """
        super().__init__()
        self.transform = transform
        self.loader = loader
        self.is_valid_file = is_valid_file
        self.image_files = get_image_files(root, self.is_valid_file)

    def __len__(self) -> int:
        return len(self.image_files)

    def __getitem__(self, index: int):
        image = self.loader(self.image_files[index])
        if self.transform is not None:
            image = self.transform(image)
        return image


def register_dataset(builder: DatasetBuilder) -> DatasetBuilder:
    """A decorator to register a dataset builder.

    The lowercase name of the builder is used in the registry.

    Args:
        builder: A dataset builder.

    Returns:
        The input dataset builder.
    """
    DATASET_REGISTRY[builder.__name__.lower()] = builder
    return builder


@register_dataset
def image_folder(
    root: str,
    subset: Optional[Union[str, List[str]]] = None,
    image_transform: Optional[ImageTransform] = None,
    **kwargs,
) -> Dataset:
    """An image folder dataset builder.

    Args:
        root: The path to the image folder dataset.
        subset: The subset(s) to use.
        image_transform: The image transformations to apply.

    Returns:
        An `ImageFolder` dataset.
    """
    if isinstance(subset, str):
        return _ImageFolder(os.path.join(root, subset), transform=image_transform)
    else:
        if subset is None:
            subsets = [
                x
                for x in os.listdir(root)
                if not x.startswith(".") and os.path.isdir(os.path.join(root, x))
            ]
        else:
            subsets = subset

        dsts = [
            _ImageFolder(os.path.join(root, s), transform=image_transform)
            for s in subsets
        ]

        return ConcatDataset(dsts)


@register_dataset
def pseudo(
    root: str,
    subset: Optional[Union[str, List[str]]] = None,
    image_transform: Optional[ImageTransform] = None,
    **kwargs,
) -> Dataset:
    """An image folder dataset builder.

    Args:
        root: The path to the image folder dataset.
        subset: The subset(s) to use.
        image_transform: The image transformations to apply.

    Returns:
        An `ImageFolder` dataset.
    """
    assert subset is None, "psuedo-labeled datasets don't have subsets"
    return PseudoDataset(
        root=root,
        transform=image_transform,
        **kwargs,
    )


@register_dataset
def concat(
    root: str,
    *,
    subset: Optional[Union[str, List[str]]] = None,
    image_transform: Optional[ImageTransform] = None,
    datasets: List[Dataset],
    **kwargs,
) -> Dataset:
    """An image folder dataset builder.

    Args:
        root: This is ignored for concat datasets.
        subset: The subset(s) to use.
        image_transform: The image transformations to apply.

    Returns:
        An `ImageFolder` dataset.
    """
    assert subset is None, "concat datasets don't support subsets"
    return ConcatDataset(datasets)


def _cifar(
    dataset_cls: Union[Type[CIFAR10], Type[CIFAR100]],
    root: str,
    subset: Optional[Union[str, List[str]]] = None,
    image_transform: Optional[ImageTransform] = None,
    **kwargs,
) -> Dataset:
    """A CIFAR dataset builder.

    Args:
        dataset_cls: One of (CIFAR10, CIFAR100).
        root: The path to the image folder dataset.
        subset: The subset(s) to use. Subets include (train, test).
        image_transform: The image transformations to apply.

    Returns:
        An `ImageFolder` dataset.
    """
    if isinstance(subset, str):
        assert subset in {
            "train",
            "test",
        }, f"invalid subset {subset}, only support {{train, test}}"
        train = subset == "train"
        return dataset_cls(
            root,
            train=train,
            download=True,
            transform=image_transform,
        )
    else:
        if subset is None:
            subsets = ["train", "test"]
        else:
            subsets = subset

        for s in subsets:
            assert s in {
                "train",
                "test",
            }, f"invalid subset {subset}, only support {{train, test}}"

        dsts = [
            dataset_cls(
                root,
                train=(s == "train"),
                download=True,
                transform=image_transform,
            )
            for s in subsets
        ]

        return ConcatDataset(dsts)


@register_dataset
def cifar10(
    root: str,
    subset: Optional[Union[str, List[str]]] = None,
    image_transform: Optional[ImageTransform] = None,
    **kwargs,
) -> Dataset:
    """A CIFAR-10 dataset builder.

    Args:
        root: The path to the image folder dataset.
        subset: The subset(s) to use. Subets include (train, test).
        image_transform: The image transformations to apply.

    Returns:
        An `ImageFolder` dataset.
    """
    return _cifar(
        CIFAR10, root, subset=subset, image_transform=image_transform, **kwargs
    )


@register_dataset
def cifar100(
    root: str,
    subset: Optional[Union[str, List[str]]] = None,
    image_transform: Optional[ImageTransform] = None,
    **kwargs,
) -> Dataset:
    """A CIFAR-100 dataset builder.

    Args:
        root: The path to the image folder dataset.
        subset: The subset(s) to use. Subets include (train, test).
        image_transform: The image transformations to apply.

    Returns:
        An `ImageFolder` dataset.
    """
    return _cifar(
        CIFAR100, root, subset=subset, image_transform=image_transform, **kwargs
    )


@register_dataset
def stl10(
    root: str,
    subset: Optional[Union[str, List[str]]] = None,
    image_transform: Optional[ImageTransform] = None,
    **kwargs,
) -> Dataset:
    """An STL10 dataset builder.

    Args:
        root: The path to the image folder dataset.
        subset: The subset(s) to use. Subets include
            (train, test, unlabeled, train+unlabeled).
        image_transform: The image transformations to apply.

    Returns:
        An `ImageFolder` dataset.
    """
    if isinstance(subset, str):
        assert subset in {
            "train",
            "test",
            "unlabeled",
            "train+unlabeled",
        }, f"invalid subset {subset}, only support {{train, test, unlabeled, train+unlabeled}}"
        return STL10(
            root,
            split=subset,
            download=True,
            transform=image_transform,
            **kwargs,
        )
    else:
        if subset is None:
            subsets = ["train", "test", "unlabeled"]
        else:
            subsets = subset

        for s in subsets:
            assert s in {
                "train",
                "test",
                "unlabeled",
                "train+unlabeled",
            }, f"invalid subset {subset}, only support {{train, test, unlabeled, train+unlabeled}}"

        dsts = [
            STL10(
                root,
                split=s,
                download=True,
                transform=image_transform,
                **kwargs,
            )
            for s in subsets
        ]

        return ConcatDataset(dsts)


@register_dataset
def image_file(
    root: str,
    subset: Optional[Union[str, List[str]]] = None,
    image_transform: Optional[ImageTransform] = None,
    **kwargs,
) -> Dataset:
    """An image folder dataset builder.

    Args:
        root: The path to the image folder dataset.
        subset: The subset(s) to use.
        image_transform: The image transformations to apply.

    Returns:
        An `ImageFolder` dataset.
    """
    if isinstance(subset, str):
        return ImageFile(os.path.join(root, subset), transform=image_transform)
    else:
        if subset is None:
            subsets = [
                x
                for x in os.listdir(root)
                if not x.startswith(".") and os.path.isdir(os.path.join(root, x))
            ]
        else:
            subsets = subset

        dsts = [
            ImageFile(os.path.join(root, s), transform=image_transform) for s in subsets
        ]

        return ConcatDataset(dsts)


def dataset(
    name: str,
    root: str,
    subset: Optional[Union[str, List[str]]] = None,
    image_transform: Optional[Callable] = None,
    **kwargs,
) -> Union[Dataset, IterableDataset]:
    """Builds a dataset.

    Args:
        name: The name of the dataset to build.
        root: The path to the image folder dataset.
        subset: The subset(s) to use.
        image_transform: The image transformations to apply.
        **kwargs: Any other arguments to forward to the underlying dataset builder.

    Returns:
        The dataset.
    """
    return DATASET_REGISTRY[name](
        to_absolute_path(root), subset=subset, image_transform=image_transform, **kwargs
    )



#
# Copyright 2021, Flyreel. All Rights Reserved.
# =============================================#
"""The trainer CLI's main function and configuration."""
import hydra
import logging
import pytorch_lightning as pl

from typing import Any, Dict
from omegaconf import OmegaConf
from hydra.utils import instantiate
from hydra.core.config_store import ConfigStore
from hydra.utils import to_absolute_path
from math import ceil

from ..config.resolvers import SESEMIConfigAttributes
from ..config.structs import SESEMIBaseConfig, ClassifierConfig, RunMode
from ..datamodules import SESEMIDataModule
from ..utils import compute_num_gpus, copy_config, load_checkpoint

logger = logging.getLogger(__name__)


config_store = ConfigStore.instance()
config_store.store(
    name="sesemi", node=SESEMIBaseConfig, group="base", package="_global_"
)
config_store.store(name="classifier", node=ClassifierConfig, group="learner")


@hydra.main(config_path="conf", config_name="/base/config")
def open_sesemi(config: SESEMIBaseConfig):
    """The trainer's main function.

    Args:
        config: The trainer config.
    """
    random_seed = pl.seed_everything(config.run.seed)

    sesemi_config_attributes = SESEMIConfigAttributes()
    OmegaConf.register_new_resolver("sesemi", sesemi_config_attributes, replace=True)

    # Expose trainer-specific attributes to config
    num_gpus = compute_num_gpus(config.run.gpus)
    sesemi_config_attributes.num_gpus = num_gpus
    sesemi_config_attributes.num_nodes = config.run.num_nodes or 1

    # Build data loaders
    accelerator = None
    if sesemi_config_attributes.num_gpus > 0:
        accelerator = config.run.accelerator or "ddp"
        assert accelerator in {"dp", "ddp"}, f"Unsupport accelerator {accelerator}"

    datamodule = SESEMIDataModule(
        config.data,
        accelerator,
        num_gpus,
        config.run.data_root,
        batch_size_per_gpu=config.run.batch_size_per_gpu,
        random_seed=random_seed,
    )

    # Expose data-specific attributes to config
    sesemi_config_attributes.iterations_per_epoch = None
    sesemi_config_attributes.max_iterations = None
    if config.data.train is not None:
        sesemi_config_attributes.iterations_per_epoch = max(
            len(x) // datamodule.train_batch_sizes_per_iteration[k]
            if config.data.train[k].drop_last
            else ceil(len(x) / datamodule.train_batch_sizes_per_iteration[k])
            for k, x in datamodule.train.items()
        )

        if config.run.num_iterations is not None:
            assert config.run.num_epochs is None
            sesemi_config_attributes.max_iterations = config.run.num_iterations
        elif config.run.num_epochs is not None:
            assert config.run.num_iterations is None
            sesemi_config_attributes.max_iterations = (
                config.run.num_epochs * sesemi_config_attributes.iterations_per_epoch
            )

    learner = instantiate(config.learner, _recursive_=False)

    trainer_config = copy_config(config.trainer) if config.trainer is not None else {}
    trainer_config["callbacks"] = [
        instantiate(c) for c in trainer_config.get("callbacks", [])
    ]

    trainer = pl.Trainer(
        **trainer_config,
        max_epochs=config.run.num_epochs,
        max_steps=config.run.num_iterations,
        accelerator=accelerator,
        gpus=config.run.gpus,
        num_nodes=config.run.num_nodes,
        resume_from_checkpoint=to_absolute_path(config.run.resume_from_checkpoint)
        if config.run.resume_from_checkpoint
        else None,
        replace_sampler_ddp=False,
    )

    if not config.run.resume_from_checkpoint and config.run.pretrained_checkpoint_path:
        load_checkpoint(
            learner, to_absolute_path(config.run.pretrained_checkpoint_path)
        )

    if config.run.mode == RunMode.FIT:
        trainer.fit(learner, datamodule)
    elif config.run.mode == RunMode.VALIDATE:
        trainer.validate(learner, datamodule.val_dataloader())
    else:
        assert config.run.mode == RunMode.TEST
        trainer.validate(learner, datamodule.test_dataloader())



#
# Copyright 2021, Flyreel. All Rights Reserved.
# =============================================#
"""Interface definition for backbones."""
import torch.nn as nn

from torch import Tensor


class Head(nn.Module):
    """The interface for image classification heads.

    Attributes:
        in_features: The number of inputs features to the head.
        out_features: The number of output features from the head.
    """

    in_features: int
    out_features: int


class LinearHead(Head):
    def __init__(self, in_features: int, out_features: int):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.lin = nn.Linear(in_features, out_features)

    def forward(self, inputs: Tensor) -> Tensor:
        return self.lin(inputs)



#
# Copyright 2021, Flyreel. All Rights Reserved.
# =============================================#
"""Interface definition for backbones."""
import torch.nn as nn


class Backbone(nn.Module):
    """The interface for image classification backbones.

    Attributes:
        out_features: The number of output features generated by the backbone.
    """

    out_features: int

    def freeze(self):
        """Freezes the backbone's parameters."""
        for m in self.modules():
            m.eval()
            for param in m.parameters():
                param.requires_grad = False



#
# Copyright 2021, Flyreel. All Rights Reserved.
# =============================================#
"""Image transforms."""
import os
import numpy as np

from PIL import ImageFilter
from typing import Any, Callable, List, Optional, Tuple, Union

from torch import Tensor
from PIL import ImageFilter
from typing import Callable, Tuple

import torch
import torchvision.transforms.functional as TF
import torchvision.transforms as T

from omegaconf import ListConfig
from torchvision import datasets, transforms

from .collation import RotationTransformer, JigsawTransformer
from .utils import validate_paths


IMAGENET_CHANNEL_MEAN = (0.485, 0.456, 0.406)
IMAGENET_CHANNEL_STD = (0.229, 0.224, 0.225)

CIFAR_CHANNEL_MEAN = (0.4914, 0.4822, 0.4465)
CIFAR_CHANNEL_STD = (0.2023, 0.1994, 0.2010)

modes_mapping = {
    "nearest": TF.InterpolationMode.NEAREST,
    "bilinear": TF.InterpolationMode.BILINEAR,
    "bicubic": TF.InterpolationMode.BICUBIC,
    "box": TF.InterpolationMode.BOX,
    "hamming": TF.InterpolationMode.HAMMING,
    "lanczos": TF.InterpolationMode.LANCZOS,
}


class GammaCorrection:
    def __init__(self, gamma_range: Tuple[float, float] = (0.5, 2.0)):
        """Initializes the gamma correction augmentation.

        Args:
            gamma_range: A tuple defining the lower and upper bound of the range.
        """
        self.gamma_range = gamma_range

    def __call__(self, x: Tensor) -> Tensor:
        """Applies random gamma correction to the input image.

        Args:
            x: The input PIL image or tensor.

        Returns:
            The gamma corrected image.
        """
        gamma = np.random.uniform(*self.gamma_range)
        return TF.adjust_gamma(x, gamma, gain=1)

    def __repr__(self) -> str:
        return self.__class__.__name__ + "(r={})".format(self.gamma_range)


class GaussianBlur:
    """Gaussian blur augmentation in SimCLR
    https://arxiv.org/abs/2002.05709
    """

    def __init__(self, sigma_range: Tuple[float, float] = (0.1, 2.0)):
        """Initializes the Gaussian blur augmentation.

        Args:
            sigma_range: A tuple defining the lower and upper bound of the range.
        """
        self.sigma_range = sigma_range

    def __call__(self, x):
        """Applies random Gaussian blur to the input image.

        Args:
            x: The input PIL image.

        Returns:
            The Gaussian blurred image.
        """
        sigma = np.random.uniform(*self.sigma_range)
        return x.filter(ImageFilter.GaussianBlur(radius=sigma))

    def __repr__(self) -> str:
        return self.__class__.__name__ + "(s={})".format(self.sigma_range)


def train_transforms(
    random_resized_crop: bool = True,
    resize: int = 256,
    crop_dim: int = 224,
    scale: Tuple[float, float] = (0.2, 1.0),
    interpolation: str = "bilinear",
    gamma_range: Tuple[float, float] = (0.5, 1.5),
    sigma_range: Tuple[float, float] = (0.1, 2.0),
    p_blur: float = 0.0,
    p_grayscale: float = 0.0,
    p_hflip: float = 0.5,
    norms: Tuple[Tuple[float, float, float], Tuple[float, float, float]] = (
        IMAGENET_CHANNEL_MEAN,
        IMAGENET_CHANNEL_STD,
    ),
    p_erase: float = 0.0,
) -> Callable:
    """Builds a torchvision training transform.

    Args:
        random_resized_crop: Whether to apply random resized cropping.
        resize: The image size to resize to if random cropping is not applied.
        crop_dim: The output crop dimension.
        scale: The scale of the random crop.
        interpolation: The interpolation mode to use when resizing.
        gamma_range: The gamma correction range.
        sigma_range: The Gaussian blur range.
        p_blur: The probability of applying Gaussian blur.
        p_grayscale: The probability of applying random grayscale.
        p_hflip: The horiziontal random flip probability.
        norms: A tuple of the normalization mean and standard deviation.
        p_erase: The probability of random erasing.

    Returns:
        A torchvision transform.
    """
    interpolation = modes_mapping[interpolation]
    augmentations = [
        GammaCorrection(gamma_range),
        transforms.RandomGrayscale(p_grayscale),
        transforms.RandomApply([GaussianBlur(sigma_range)], p_blur),
        transforms.RandomHorizontalFlip(p_hflip),
        transforms.ToTensor(),
        transforms.Normalize(*norms),
        transforms.RandomErasing(p=p_erase, value="random"),
    ]
    if random_resized_crop:
        return transforms.Compose(
            [transforms.RandomResizedCrop(crop_dim, scale, interpolation=interpolation)]
            + augmentations
        )
    else:
        return transforms.Compose(
            [transforms.Resize(resize, interpolation), transforms.RandomCrop(crop_dim)]
            + augmentations
        )


class TwoViewsTransform:
    """Randomly crops two views of the same image."""

    def __init__(
        self,
        transform: Callable,
    ):
        """Initializes the two-views transform.

        Args:
            transform: The base transform.
        """
        self.transform = transform

    def __call__(self, x: Tensor) -> Tuple[Tensor, Tensor]:
        """Applies random two-views transform to the input image.

        Args:
            x: The input tensor.

        Returns:
            A tuple of two randomly cropped views with augmentations.
        """
        one = self.transform(x)
        two = self.transform(x)
        return (one, two)


class MultiViewTransform:
    """A multi-view image transform."""

    def __init__(
        self,
        num_views: int,
        image_augmentations: Optional[Union[Callable, List[Callable]]] = None,
    ):
        self.num_views = num_views

        if isinstance(image_augmentations, (list, ListConfig)):
            assert (
                len(image_augmentations) == num_views
            ), f"must provide {num_views} augmentations if multiple given"
            self.image_augmentations = image_augmentations
        else:
            self.image_augmentations = [image_augmentations] * self.num_views

    def __call__(self, image) -> Tuple[Any, ...]:
        """Generates a transformed data batch of rotation prediction data.

        Arguments:
            image: An input image.

        Returns:
            A tuple of augmented views.
        """

        views = []
        for i in range(self.num_views):
            if self.image_augmentations[i] is not None:
                view = self.image_augmentations[i](image)
            else:
                view = image
            views.append(view)
        return tuple(views)


def center_crop_transforms(
    resize: int = 256,
    crop_dim: int = 224,
    interpolation: str = "bilinear",
    norms: Tuple[Tuple[float, float, float], Tuple[float, float, float]] = (
        IMAGENET_CHANNEL_MEAN,
        IMAGENET_CHANNEL_STD,
    ),
) -> Callable:
    """Builds a center cropping transform.

    Args:
        resize: The image size to resize to if random cropping is not applied.
        crop_dim: The output crop dimension.
        interpolation: The interpolation mode to use when resizing.
        norms: A tuple of the normalization mean and standard deviation.

    Returns:
        A torchvision transform.
    """
    return transforms.Compose(
        [
            transforms.Resize(resize, modes_mapping[interpolation]),
            transforms.CenterCrop(crop_dim),
            transforms.ToTensor(),
            transforms.Normalize(*norms),
        ]
    )


def multi_crop_transforms(
    resize: int = 256,
    crop_dim: int = 224,
    num_crop: int = 5,
    interpolation: str = "bilinear",
    norms: Tuple[Tuple[float, float, float], Tuple[float, float, float]] = (
        IMAGENET_CHANNEL_MEAN,
        IMAGENET_CHANNEL_STD,
    ),
) -> Callable:
    """Builds a multi-crop transform.

    Args:
        resize: The image size to resize to if random cropping is not applied.
        crop_dim: The output crop dimension.
        num_crop: The number of crops to generate.
        interpolation: The interpolation mode to use when resizing.
        norms: A tuple of the normalization mean and standard deviation.

    Returns:
        A torchvision transform.
    """
    to_tensor = transforms.ToTensor()
    normalize = transforms.Normalize(*norms)
    Lambda = transforms.Lambda
    if num_crop == 5:
        multi_crop = transforms.FiveCrop
    elif num_crop == 10:
        multi_crop = transforms.TenCrop
    else:
        raise NotImplementedError("Number of crops should be integer of 5 or 10")

    return transforms.Compose(
        [
            transforms.Resize(resize, modes_mapping[interpolation]),
            multi_crop(crop_dim),  # this is a list of PIL Images
            Lambda(lambda crops: torch.stack([to_tensor(crop) for crop in crops])),
            Lambda(lambda crops: torch.stack([normalize(crop) for crop in crops])),
        ]
    )


def cifar_train_transforms() -> Callable:
    """Returns the standard CIFAR training transforms."""
    return T.Compose(
        [
            T.RandomResizedCrop(32),
            T.RandomHorizontalFlip(p=0.5),
            T.RandomApply([T.ColorJitter(0.4, 0.4, 0.4, 0.1)], p=0.8),
            T.RandomGrayscale(p=0.2),
            T.ToTensor(),
            T.Normalize(CIFAR_CHANNEL_MEAN, CIFAR_CHANNEL_STD),
        ]
    )


def cifar_test_transforms() -> Callable:
    """Returns the standard CIFAR test-time transform."""
    return T.Compose(
        [
            T.ToTensor(),
            T.Normalize(CIFAR_CHANNEL_MEAN, CIFAR_CHANNEL_STD),
        ]
    )


if __name__ == "__main__":
    import argparse
    from tqdm import trange

    parser = argparse.ArgumentParser(
        description="Dataset Visualization",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--img-dir",
        required=True,
        help="`img_dir` must have one or more subdirs containing images",
    )
    parser.add_argument("--head", default=10, type=int, help="visualize first k images")
    parser.add_argument(
        "--hflip", action="store_true", help="apply random horizontal flip"
    )
    parser.add_argument("--erase", action="store_true", help="apply random erase")
    parser.add_argument(
        "--gamma",
        action="store_true",
        help="apply random luminance and gamma correction",
    )
    parser.add_argument(
        "--normalize",
        action="store_true",
        help="apply channel-wise mean-std normalization",
    )
    parser.add_argument(
        "--visualization",
        choices=["none", "rotation", "jigsaw"],
        default="rotation",
        help="visualize transformations on unlabeled data",
    )
    parser.add_argument(
        "--out-dir",
        default="./sample_dataset_vis",
        help="directory to save images for visualization",
    )
    args = parser.parse_args()

    validate_paths([args.img_dir])
    os.makedirs(args.out_dir, exist_ok=True)

    print("args:")
    for key, val in args.__dict__.items():
        print("    {:20} {}".format(key, val))

    p_hflip = 0.5 if args.hflip else 0.0
    p_erase = 0.5 if args.erase else 0.0
    gamma_range = (0.5, 1.5) if args.gamma else (1.0, 1.0)
    (mean, std) = (
        (IMAGENET_CHANNEL_MEAN, IMAGENET_CHANNEL_STD)
        if args.normalize
        else ((0.0, 0.0, 0.0), (1.0, 1.0, 1.0))
    )

    transformations = train_transforms(
        gamma_range=gamma_range, p_hflip=p_hflip, norms=(mean, std), p_erase=p_erase
    )
    dataset = datasets.ImageFolder(args.img_dir, transformations)
    print("transformations:\n", transformations)
    print("dataset size: {}".format(len(dataset)))
    to_pil_image = transforms.ToPILImage()
    rotate = RotationTransformer()
    jigsaw = JigsawTransformer(p_grayscale=0.5)
    for i in trange(args.head):
        fpath = dataset.imgs[i][0]
        fname = fpath.split("/")[-1]
        x, dummy_label = dataset[i]
        if args.visualization == "none":
            image = to_pil_image(x)
            image.save(os.path.join(args.out_dir, fname))
            continue
        if args.visualization == "rotation":
            tensors, indices = rotate([(x, dummy_label)])
        elif args.visualization == "jigsaw":
            tensors, indices = jigsaw([(x, dummy_label)])
        for x, ind in zip(*(tensors, indices)):
            image = to_pil_image(x)
            image.save(
                os.path.join(args.out_dir, f"vis_{args.visualization}_{ind}_" + fname)
            )



#
# Copyright 2021, Flyreel. All Rights Reserved.
# =============================================#
"""Weight schedulers."""
from ..utils import sigmoid_rampup


class WeightScheduler:
    """The interface for weight schedulers."""

    def __call__(self, step: int) -> float:
        """Computes the weight for the given step number.

        Attributes:
            step: The step number.

        Returns:
            The weight.
        """
        raise NotImplementedError()


class SigmoidRampupScheduler(WeightScheduler):
    """A sigmoid ramp-up weight scheduler."""

    def __init__(self, weight: float, stop_rampup: int):
        """Initializes the weight scheduler.

        Args:
            weight: The weight to apply to the sigmoidal function.
            stop_rampup: The step to stop ramping up the output weight.
        """
        self.weight = weight
        self.stop_rampup = stop_rampup

    def __call__(self, step: int) -> float:
        return self.weight * sigmoid_rampup(step, self.stop_rampup)



#
# Copyright 2021, Flyreel. All Rights Reserved.
# =============================================#
"""PyTorch Lightning data modules."""
import pytorch_lightning as pl
import os.path as osp
import logging

from math import ceil
from torch.utils.data.dataset import Dataset
from typing import Dict, List, Optional, Union
from torch.utils.data import DataLoader, DistributedSampler
from hydra.utils import instantiate, to_absolute_path

from sesemi.utils import copy_config

from .config.structs import DataConfig, DataLoaderConfig, DatasetConfig

logger = logging.getLogger(__name__)


class SESEMIDataModule(pl.LightningDataModule):
    """The main SESEMI data module.

    Attributes:
        train (Optional[Dict[str, Dataset]]): An optional dictionary of training datasets.
        val (Optional[Dataset]): An optional validation dataset.
        test (Optional[Dataset]): An optional test dataset.
        train_batch_sizes (Dict[str, int]): The batch size used for each training data loader.
        train_batch_sizes_per_gpu (Dict[str, int]): The batch size used per GPU for each training
            data loader.
        train_batch_sizes_per_iteration (Dict[str, int]): The batch size per iteration for each
            training data loader.
    """

    def __init__(
        self,
        config: DataConfig,
        accelerator: Optional[str],
        num_gpus: int,
        data_root: str,
        batch_size_per_gpu: Optional[int] = None,
        random_seed: int = 0,
    ):
        """Initializes the data module.

        Args:
            config: The data config.
            accelerator: The accelerator being used.
            num_gpus: The number of GPUs being used.
            data_root: The data root to look for datasets with relative paths.
            batch_size_per_gpu: An optional default batch size per GPU to use with data loaders.
            random_seed: The random seed to initialize DDP data loaders.
        """
        super().__init__()
        self.config = config
        self.accelerator = accelerator
        self.num_gpus = num_gpus
        self.data_root = data_root
        self.random_seed = random_seed
        self.batch_size_per_gpu = batch_size_per_gpu
        self._build()

    def _build_dataset(self, config: DatasetConfig) -> Dataset:
        if config.root is None:
            dataset_root = to_absolute_path(self.data_root)
        elif osp.isabs(config.root):
            dataset_root = config.root
        else:
            dataset_root = to_absolute_path(osp.join(self.data_root, config.root))

        dataset_config = copy_config(config)
        dataset_config.pop("root")

        return instantiate(dataset_config, root=dataset_root)

    def _build(self):
        self.train, self.val, self.test = None, None, None
        self.train_batch_sizes = {}
        self.train_batch_sizes_per_gpu = {}
        self.train_batch_sizes_per_iteration = {}
        if self.config.train is not None:
            self.train = {
                key: self._build_dataset(value.dataset)
                for key, value in self.config.train.items()
            }

            for key, value in self.config.train.items():
                if value.batch_size is not None:
                    assert value.batch_size_per_gpu is None
                    if self.accelerator == "dp":
                        batch_size_per_gpu = ceil(value.batch_size / self.num_gpus)
                    else:
                        batch_size_per_gpu = value.batch_size
                elif value.batch_size_per_gpu is not None:
                    batch_size_per_gpu = value.batch_size_per_gpu
                elif self.batch_size_per_gpu is not None:
                    batch_size_per_gpu = self.batch_size_per_gpu
                else:
                    batch_size_per_gpu = 1

                self.train_batch_sizes_per_gpu[key] = batch_size_per_gpu

            self.train_batch_sizes_per_iteration = {
                key: self.train_batch_sizes_per_gpu[key] * max(self.num_gpus, 1)
                for key in self.config.train.keys()
            }

            self.train_batch_sizes = {
                key: self.train_batch_sizes_per_gpu[key]
                if self.accelerator == "ddp"
                else self.train_batch_sizes_per_iteration[key]
                for key in self.config.train.keys()
            }

        if self.config.val is not None:
            self.val = self._build_dataset(self.config.val.dataset)

        if self.config.test is not None:
            self.test = self._build_dataset(self.config.test.dataset)

    def train_dataloader(
        self,
    ) -> Union[DataLoader, List[DataLoader], Dict[str, DataLoader]]:
        assert self.train is not None
        assert self.config.train is not None
        assert self.train_batch_sizes is not None

        train_dataloaders = {}
        for key, dataset in self.train.items():
            dataset_config = copy_config(self.config.train[key])
            dataset_config.pop("dataset")
            dataset_config.pop("batch_size")
            dataset_config.pop("batch_size_per_gpu")

            if self.accelerator == "ddp":
                dataset_config.pop("shuffle")
                train_dataloaders[key] = instantiate(
                    dataset_config,
                    dataset=dataset,
                    batch_size=self.train_batch_sizes[key],
                    sampler=DistributedSampler(
                        dataset,
                        shuffle=self.config.train[key].shuffle,
                        seed=self.random_seed,
                    ),
                )
            else:
                train_dataloaders[key] = instantiate(
                    dataset_config,
                    dataset=dataset,
                    batch_size=self.train_batch_sizes[key],
                )

        return train_dataloaders

    def _build_evaluation_data_loader(
        self, config: DataLoaderConfig, dataset: Dataset
    ) -> DataLoader:
        dataset_config = copy_config(config)
        dataset_config.pop("dataset")
        dataset_config.pop("batch_size")
        dataset_config.pop("batch_size_per_gpu")

        assert (
            config.batch_size is None or config.batch_size_per_gpu is None
        ), "cannot set both batch_size and batch_size_per_gpu"

        if self.accelerator == "dp":
            if config.batch_size is not None:
                batch_size = config.batch_size
            elif config.batch_size_per_gpu is not None:
                batch_size = config.batch_size_per_gpu * self.num_gpus
            elif self.batch_size_per_gpu is not None:
                batch_size = self.batch_size_per_gpu * self.num_gpus
            else:
                batch_size = 1
        else:
            batch_size = (
                config.batch_size
                or config.batch_size_per_gpu
                or self.batch_size_per_gpu
                or 1
            )

        return instantiate(dataset_config, dataset=dataset, batch_size=batch_size)

    def val_dataloader(self) -> Union[DataLoader, List[DataLoader]]:
        assert self.val is not None
        assert self.config.val is not None
        return self._build_evaluation_data_loader(self.config.val, self.val)

    def test_dataloader(self) -> Union[DataLoader, List[DataLoader]]:
        assert self.test is not None
        assert self.config.test is not None
        return self._build_evaluation_data_loader(self.config.test, self.test)



#
# Copyright 2021, Flyreel. All Rights Reserved.
# =============================================#
"""Backbones from the rwightman/pytorch-image-models repository."""
import logging

from .base import Backbone
from ..utils import load_torch_hub_model

PYTORCH_IMAGE_MODELS_REPO = "rwightman/pytorch-image-models"

logger = logging.getLogger(__name__)


SUPPORTED_BACKBONES = (
    # The following backbones strike a balance between accuracy and model size, with optional
    # pretrained ImageNet weights. For a summary of their ImageNet performance, see
    # <https://github.com/rwightman/pytorch-image-models/blob/master/results/results-imagenet.csv>.
    # Compared with the defaults, the "d" variants (e.g., resnet50d, resnest50d)
    # replace the 7x7 conv in the input stem with three 3x3 convs.
    # And in the downsampling block, a 2x2 avg_pool with stride 2 is added before conv,
    # whose stride is changed to 1. Described in `Bag of Tricks <https://arxiv.org/abs/1812.01187>`.
    # ResNet models.
    "resnet18",
    "resnet18d",
    "resnet34",
    "resnet34d",
    "resnet50",
    "resnet50d",
    "resnet101d",
    "resnet152d",
    # ResNeXt models.
    "resnext50_32x4d",
    "resnext50d_32x4d",
    "resnext101_32x8d",
    # Squeeze and Excite models.
    "seresnet50",
    "seresnet152d",
    "seresnext26d_32x4d",
    "seresnext26t_32x4d",
    "seresnext50_32x4d",
    # ResNeSt models.
    "resnest14d",
    "resnest26d",
    "resnest50d",
    "resnest101e",
    "resnest200e",
    "resnest269e",
    "resnest50d_1s4x24d",
    "resnest50d_4s2x40d",
    # ResNet-RS models.
    "resnetrs50",
    "resnetrs101",
    "resnetrs152",
    "resnetrs200",
    # DenseNet models.
    "densenet121",
    "densenet169",
    "densenet201",
    "densenet161",
    # Inception models.
    "inception_v3",
    "inception_v4",
    "inception_resnet_v2",
    # Xception models.
    "xception",
    "xception41",
    "xception65",
    "xception71",
    # EfficientNet models.
    "tf_efficientnet_b0",
    "tf_efficientnet_b1",
    "tf_efficientnet_b2",
    "tf_efficientnet_b3",
    "tf_efficientnet_b4",
    "tf_efficientnet_b5",
    "tf_efficientnet_b6",
    "tf_efficientnet_b7",
    # EfficientNet models trained with noisy student.
    "tf_efficientnet_b0_ns",
    "tf_efficientnet_b1_ns",
    "tf_efficientnet_b2_ns",
    "tf_efficientnet_b3_ns",
    "tf_efficientnet_b4_ns",
    "tf_efficientnet_b5_ns",
    "tf_efficientnet_b6_ns",
    "tf_efficientnet_b7_ns",
)


class PyTorchImageModels(Backbone):
    def __init__(
        self,
        name: str = "resnet50d",
        pretrained: bool = True,
        global_pool: str = "avg",
        drop_rate: float = 0.0,
        freeze: bool = False,
        **kwargs,
    ):
        """Builds the pytorch-image-models backbone.

        Args:
            name: The name of the backbone.
            pretrained: Whether to load the default pretrained model weights.
            global_pool: The kind of pooling to use. Can be one of (avg, max, avgmax, catavgmax).
            drop_rate: The dropout rate.
            freeze: Whether to freeze the backbone's weights.
        """
        if name not in SUPPORTED_BACKBONES:
            logger.warn(
                f"backbone {name} is not one of the supported backbones: {SUPPORTED_BACKBONES}"
            )

        super().__init__()
        self.encoder = load_torch_hub_model(
            PYTORCH_IMAGE_MODELS_REPO,
            name,
            pretrained,
            num_classes=0,
            global_pool=global_pool,
            drop_rate=drop_rate,
            **kwargs,
        )

        self.out_features = self.encoder.num_features
        if global_pool == "catavgmax":
            self.out_features *= 2

        if freeze:
            self.freeze()

    def forward(self, x):
        return self.encoder(x)



#
# Copyright 2021, Flyreel. All Rights Reserved.
# =============================================#
"""Loss heads."""
import torch
import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor
from typing import Dict, Any, Optional
from pytorch_lightning.loggers.base import LightningLoggerBase

from ..backbones.base import Backbone
from ..heads.base import Head
from ...losses import (
    softmax_mse_loss,
    kl_div_loss,
)


class LossHead(nn.Module):
    """The interface for loss heads.

    Attributes:
        logger: An optional PyTorch Lightning logger.
    """

    def __init__(
        self,
        logger: Optional[LightningLoggerBase] = None,
    ):
        """Initializes the loss head.

        Args:
            logger: An optional PyTorch Lightning logger.
        """
        super().__init__()
        self.logger = logger

    def build(self, backbones: Dict[str, Backbone], heads: Dict[str, Head], **kwargs):
        """Builds the loss head.

        Args:
            backbones: A dictionary of shared backbones. During the build, new backbones can be
                added. As this is actually an `nn.ModuleDict` which tracks the parameters,
                these new backbones should not be saved to the loss head object to avoid
                double-tracking.
            heads: A dictionary of shared heads similar to the backbones
        """

    def forward(
        self,
        data: Dict[str, Any],
        backbones: Dict[str, Backbone],
        heads: Dict[str, Head],
        features: Dict[str, Any],
        step: int,
        **kwargs,
    ) -> Tensor:
        """Computes the loss.

        Args:
            data: A dictionary of data batches.
            backbones: A dictionary of shared backbones. This should not be altered.
            heads: A dictionary of shared heads. This should not be altered.
            features: A dictionary of shared features. Additional tensors can be added to this.
            step: The training step number.
            **kwargs: Placeholder for other arguments that may be added.
        """
        raise NotImplementedError()


class RotationPredictionLossHead(LossHead):
    """The rotation prediction loss head.
    https://arxiv.org/abs/1803.07728
    """

    def __init__(
        self,
        input_data: str,
        input_backbone: str = "backbone",
        num_pretext_classes: int = 4,
        logger: Optional[LightningLoggerBase] = None,
    ):
        """Initializes the loss head.

        Args:
            input_data: The key used to get the rotation prediction input data.
            input_backbone: The key used to get the backbone for feature extraction.
            num_pretext_classes: Number of pretext labels.
            logger: An optional PyTorch Lightning logger.
        """
        super().__init__(logger)
        self.input_data = input_data
        self.input_backbone = input_backbone
        self.num_pretext_classes = num_pretext_classes

    def build(self, backbones: Dict[str, Backbone], heads: Dict[str, Head], **kwargs):
        self.fc_unsupervised = nn.Linear(
            backbones[self.input_backbone].out_features, self.num_pretext_classes
        )

    def forward(
        self,
        data: Dict[str, Any],
        backbones: Dict[str, Backbone],
        heads: Dict[str, Head],
        features: Dict[str, Any],
        step: int,
        **kwargs,
    ) -> Tensor:
        inputs, targets = data[self.input_data]
        feats = backbones[self.input_backbone](inputs)
        logits = self.fc_unsupervised(feats)
        loss_u = F.cross_entropy(logits, targets, reduction="none")
        return loss_u


class JigsawPredictionLossHead(RotationPredictionLossHead):
    """The jigsaw prediction loss head. Idea and implementation are adapted from
    https://arxiv.org/abs/1903.06864
    """

    def __init__(
        self,
        input_data: str,
        input_backbone: str = "backbone",
        num_pretext_classes: int = 6,
        logger: Optional[LightningLoggerBase] = None,
    ):
        super(JigsawPredictionLossHead, self).__init__(
            input_data, num_pretext_classes=num_pretext_classes
        )


class EntropyMinimizationLossHead(LossHead):
    """The entropy minimization loss head.
    https://papers.nips.cc/paper/2004/file/96f2b50b5d3613adf9c27049b2a888c7-Paper.pdf
    """

    def __init__(
        self,
        input_data: str,
        input_backbone: str = "backbone",
        predict_head: str = "supervised",
        logger: Optional[LightningLoggerBase] = None,
    ):
        """Initializes the loss head.

        Args:
            input_data: The key used to get the unlabeled input data.
            input_backbone: The key used to get the backbone for feature extraction.
            predict_head: The prediction module used to compute output logits.
            logger: An optional PyTorch Lightning logger.
        """
        super().__init__(logger)
        self.input_data = input_data
        self.input_backbone = input_backbone
        self.predict_head = predict_head

    def forward(
        self,
        data: Dict[str, Any],
        backbones: Dict[str, Backbone],
        heads: Dict[str, Head],
        features: Dict[str, Any],
        step: int,
        **kwargs,
    ) -> Tensor:
        inputs, _ = data[self.input_data]
        feats = backbones[self.input_backbone](inputs)
        logits = heads[self.predict_head](feats)
        loss_u = -F.softmax(logits, dim=-1) * F.log_softmax(logits, dim=-1)
        return loss_u


class ConsistencyLossHead(EntropyMinimizationLossHead):
    """The consistency loss head following the Pi Model.
    https://arxiv.org/abs/1610.02242
    """

    def __init__(
        self,
        input_data: str,
        input_backbone: str = "backbone",
        predict_head: str = "supervised",
        loss_fn: str = "mse",
        logger: Optional[LightningLoggerBase] = None,
    ):
        """
        Args:
            input_data: The key used to get the unlabeled input data,
                which in this case returns two views of the same image.
            input_backbone: The key used to get the backbone for feature extraction.
            predict_head: The prediction module used to compute output logits.
            loss_fn: The loss function to compute the consistency between two views.
            logger: An optional PyTorch Lightning logger.
        """
        super(ConsistencyLossHead, self).__init__(input_data)
        if loss_fn == "mse":
            self.loss_fn = softmax_mse_loss
        elif loss_fn == "kl_div":
            self.loss_fn = kl_div_loss
        else:
            raise ValueError(
                loss_fn,
                "is not a supported consistency loss function. "
                "Choose between `mse` or `kl_div`. Default `mse`.",
            )

    def forward(
        self,
        data: Dict[str, Any],
        backbones: Dict[str, Backbone],
        heads: Dict[str, Head],
        features: Dict[str, Any],
        step: int,
        **kwargs,
    ) -> Tensor:
        (view1, view2), _ = data[self.input_data]
        feats1 = backbones[self.input_backbone](view1)
        feats2 = backbones[self.input_backbone](view2)
        logits1 = heads[self.predict_head](feats1)
        logits2 = heads[self.predict_head](feats2)
        loss_u = self.loss_fn(logits1, logits2)
        return loss_u


class EMAConsistencyLossHead(ConsistencyLossHead):
    """The EMA consistency loss head following Mean Teacher.
    https://arxiv.org/abs/1703.01780
    """

    def __init__(
        self,
        input_data: str,
        student_backbone: str = "backbone",
        teacher_backbone: str = "backbone_ema",
        student_head: str = "supervised",
        teacher_head: str = "supervised_ema",
        loss_fn: str = "mse",
        logger: Optional[LightningLoggerBase] = None,
    ):
        """
        Args:
            input_data: The key used to get the unlabeled input data,
                which in this case returns two views of the same image.
            student_backbone: The student backbone for feature extraction.
            teacher_backbone: The teacher backbone for feature extraction.
            student_head: The student module used to compute output logits.
            teacher_head: The teacher module used to compute output logits.
            loss_fn: The loss function to compute the consistency between two views.
            logger: An optional PyTorch Lightning logger.
        """
        super(EMAConsistencyLossHead, self).__init__(input_data, loss_fn=loss_fn)
        self.student_backbone = student_backbone
        self.teacher_backbone = teacher_backbone
        self.student_head = student_head
        self.teacher_head = teacher_head

    def forward(
        self,
        data: Dict[str, Any],
        backbones: Dict[str, Backbone],
        heads: Dict[str, Head],
        features: Dict[str, Any],
        step: int,
        **kwargs,
    ) -> Tensor:
        (view1, view2), _ = data[self.input_data]
        student_feats = backbones[self.student_backbone](view1)
        teacher_feats = backbones[self.teacher_backbone](view2)
        student_logits = heads[self.student_head](student_feats)
        teacher_logits = heads[self.teacher_head](teacher_feats)
        loss_u = self.loss_fn(student_logits, teacher_logits)
        return loss_u


class FixMatchLossHead(LossHead):
    """The FixMatch loss head.

    @article{Sohn2020FixMatchSS,
        title={FixMatch: Simplifying Semi-Supervised Learning with Consistency and Confidence},
        author={Kihyuk Sohn and David Berthelot and Chun-Liang Li and Zizhao Zhang and Nicholas Carlini and Ekin Dogus Cubuk and Alexey Kurakin and Han Zhang and Colin Raffel},
        journal={ArXiv},
        year={2020},
        volume={abs/2001.07685}
    }
    """

    def __init__(
        self,
        data: str,
        student_backbone: str = "backbone",
        teacher_backbone: Optional[str] = None,
        student_head: str = "supervised",
        teacher_head: Optional[str] = None,
        threshold: float = 0.5,
        logger: Optional[LightningLoggerBase] = None,
    ):
        """Initializes the loss head.

        Args:
            data: The data key.
            student_backbone: The student's backbone key.
            teacher_backbone: The teacher's backbone key. Defaults to the student's.
            student_head: The student's head key.
            teacher_head: The teacher's head key. Defaults to the student's.
            threshold: The threshold used to filter low confidence predictions
                made by the teacher.
            logger: An optional PyTorch Lightning logger.
        """
        super().__init__(logger)
        self.data = data
        self.student_backbone = student_backbone
        self.teacher_backbone = teacher_backbone or student_backbone
        self.student_head = student_head
        self.teacher_head = teacher_head or student_head
        self.threshold = threshold

    def forward(
        self,
        data: Dict[str, Any],
        backbones: Dict[str, Backbone],
        heads: Dict[str, Head],
        features: Dict[str, Any],
        step: int,
        **kwargs,
    ) -> Tensor:
        weakly_augmented, strongly_augmented = data[self.data]
        student_backbone = backbones[self.student_backbone]
        student_head = heads[self.student_head]
        teacher_backbone = backbones[self.teacher_backbone]
        teacher_head = heads[self.teacher_head]

        weakly_augmented_features = teacher_backbone(weakly_augmented)
        strongly_augmented_features = student_backbone(strongly_augmented)

        weakly_augmented_logits = teacher_head(weakly_augmented_features).detach()
        strongly_augmented_logits = student_head(strongly_augmented_features)

        weakly_augmented_probs = torch.softmax(weakly_augmented_logits, dim=-1)
        weakly_augmented_labels = torch.argmax(weakly_augmented_probs, dim=-1).to(
            torch.long
        )

        loss_weight = (weakly_augmented_probs.max(dim=-1)[0] >= self.threshold).to(
            torch.float32
        )

        loss = (
            F.cross_entropy(
                strongly_augmented_logits,
                weakly_augmented_labels,
                reduction="none",
            )
            * loss_weight
        )

        total_loss_weight = torch.sum(loss_weight)
        total_loss = torch.sum(loss)

        loss = total_loss / (total_loss_weight + 1e-8)
        return loss



#
# Copyright 2021, Flyreel. All Rights Reserved.
# =============================================#
"""PyTorch collation callables."""
import torch
import torchvision
import torchvision.transforms.functional as TF

from torch import Tensor
from typing import List, Tuple, Callable


class RotationTransformer:
    """A collation callable to transform a data batch for use with a rotation prediction task.

    Input images are rotated 0, 90, 180, and 270 degrees.
    """

    def __init__(self, return_supervised_labels: bool = False):
        """Initializes the rotation collation callable.

        Args:
            return_supervised_labels: Whether to return supervised class labels or pretext labels.
        """
        self.return_supervised_labels = return_supervised_labels

    def __call__(self, batch: List[Tuple[Tensor, Tensor]]) -> Tuple[Tensor, Tensor]:
        """Generates a transformed batch of rotation prediction data.

        Args:
            batch: A tuple containing a batch of input image tensors and supervised labels.

        Returns:
            A tuple of rotated image tensors and their associated supervised or pretext labels, where:

                0  corresponds to 0 degrees.
                1  corresponds to 90 degrees.
                2  corresponds to 180 degrees.
                3  corresponds to 270 degrees.
        """
        tensors, labels, rotation_labels = [], [], []
        for tensor, label in batch:
            for k in range(4):
                if k == 0:
                    t = tensor
                else:
                    t = torch.rot90(tensor, k, dims=[1, 2])
                tensors.append(t)
                labels.append(torch.LongTensor([label]))
                rotation_labels.append(torch.LongTensor([k]))
        x = torch.stack(tensors, dim=0)
        y = torch.cat(labels, dim=0)
        p = torch.cat(rotation_labels, dim=0)
        return (x, y) if self.return_supervised_labels else (x, p)


class JigsawTransformer:
    """A collation callable to transform a data batch for use with a jigsaw prediction task.

    We select a set of P=5 patch permutations for the jigsaw task by using the maximal Hamming distance.
    https://github.com/bbrattoli/JigsawPuzzlePytorch
    """

    def __init__(
        self,
        grid_size: int = 3,
        p_grayscale: float = 0.1,
        return_supervised_labels: bool = False,
    ):
        """Initializes the jigsaw collation callable.

        Args:
            grid_size: grid of n x n patches.
            p_grayscale: probability of converting to grayscale.
            return_supervised_labels: Whether to return supervised class labels or pretext labels.
        """
        self.grid_size = grid_size
        self.num_grids = grid_size ** 2
        self.p_grayscale = p_grayscale
        self.return_supervised_labels = return_supervised_labels
        self.patch_permutations = torch.tensor(
            [
                [5, 7, 2, 6, 0, 3, 8, 1, 4],
                [0, 1, 3, 2, 4, 5, 6, 7, 8],
                [1, 0, 4, 3, 2, 6, 5, 8, 7],
                [2, 3, 0, 1, 5, 8, 7, 4, 6],
                [3, 2, 1, 0, 8, 7, 4, 6, 5],
            ]
        )
        # The number of jigsaw labels is P + 1, with the last label
        # indicating no patch permutation (no shuffling) applied.
        self.num_jigsaw_labels = self.patch_permutations.size(0) + 1

    def crop_patch(self, x: Tensor, idx: int) -> Tensor:
        """Crops a patch of equal height and width dimensions from an input tensor.

        Args:
            x: the input tensor.
            idx: the location index of the ith patch.

        Returns:
            A torchvision transformed tensor.
        """
        window = int(x.size(1) / self.grid_size)
        row = int(idx / self.grid_size)
        col = idx % self.grid_size
        patch = TF.crop(x, row * window, col * window, window, window)
        return self.transform()(patch)

    def transform(self) -> Callable:
        """Image transformations.

        Returns:
            Torchvision transform.
        """
        return torchvision.transforms.Compose(
            [torchvision.transforms.RandomGrayscale(self.p_grayscale)]
        )

    def __call__(self, batch: List[Tuple[Tensor, Tensor]]) -> Tuple[Tensor, Tensor]:
        """Generates a transformed batch of jigsaw prediction data.

        Args:
            batch: A tuple containing a batch of input image tensors and supervised labels.

        Returns:
            A tuple of jigsaw image tensors and their associated supervised or pretext labels.
        """
        b = len(batch) * (self.num_jigsaw_labels)
        c, h, w = batch[0][0].size()
        tensors, labels, jigsaw_labels = [], [], []
        for tensor, label in batch:
            patches = torch.stack(
                [self.crop_patch(tensor, idx) for idx in range(self.num_grids)], dim=0
            )
            for perm_idx in range(self.num_jigsaw_labels):
                if perm_idx == self.num_jigsaw_labels - 1:
                    data = patches
                else:
                    shuffle_order = self.patch_permutations[perm_idx]
                    data = patches[shuffle_order]
                data = torchvision.utils.make_grid(data, self.grid_size, padding=0)
                tensors.append(data)
                labels.append(torch.LongTensor([label]))
                jigsaw_labels.append(torch.LongTensor([perm_idx]))
        tensors = torch.stack(tensors, dim=0)
        x = torch.zeros(b, c, h, w)
        x[..., : tensors.size(2), : tensors.size(3)] = tensors
        y = torch.cat(labels, dim=0)
        p = torch.cat(jigsaw_labels, dim=0)
        return (x, y) if self.return_supervised_labels else (x, p)



#
# Copyright 2021, Flyreel. All Rights Reserved.
# =============================================#
"""Learning rate schedulers."""
from typing import Optional
import warnings
from torch.optim import Optimizer

from torch.optim.lr_scheduler import _LRScheduler

import logging

logger = logging.getLogger(__name__)


class PolynomialLR(_LRScheduler):
    """A polynomial learning rate scheduler."""

    def __init__(
        self,
        optimizer: Optimizer,
        warmup_lr: float,
        lr_pow: float,
        max_iters: int,
        warmup_iters: Optional[int] = None,
        warmup_epochs: Optional[int] = None,
        iters_per_epoch: Optional[int] = None,
        last_epoch: int = -1,
    ):
        """Initializes the polynomial learning rate scheduler.

        Args:
            optimizer: The optimizer.
            warmup_lr: The initial warmup learning rate.
            lr_pow: The power of the polynomial update rule.
            max_iters: The maximum number of updates.
            warmup_iters: The number of iterations to warmup. Mutually exclusive with
                `warmup_epochs`.
            warmup_epochs: The number of warmup epochs. Mutually exclusive with `warmup_iters`.
            iters_per_epoch: The number of iterations per epoch. Must be set if `warmup_epochs` is
                set.
            last_epoch: The index of the last epoch.
        """
        if warmup_iters is None:
            assert (
                warmup_epochs is not None and iters_per_epoch is not None
            ), "must specify either warmup_iters or warmup_epochs and iters_per_epoch"

            self.iters_per_epoch = iters_per_epoch
            self.warmup_iters = warmup_epochs * iters_per_epoch
        else:
            self.warmup_iters = warmup_iters

        logger.info(f"Using warmup iterations of {self.warmup_iters}")

        self.warmup_lr = warmup_lr
        self.lr_pow = lr_pow
        self.max_iters = max_iters
        assert self.max_iters >= self.warmup_iters
        super().__init__(optimizer, last_epoch)  # type: ignore

    def get_lr(self):
        if not self._get_lr_called_within_step:
            warnings.warn(
                "To get the last learning rate computed by the scheduler, "
                "please use `get_last_lr()`.",
                UserWarning,
            )

        if self.last_epoch >= self.max_iters:
            return [0.0 for _ in self.optimizer.param_groups]
        else:
            if self.last_epoch < self.warmup_iters:
                frac = self.last_epoch / self.warmup_iters
                return [
                    self.warmup_lr + (group["initial_lr"] - self.warmup_lr) * frac
                    for group in self.optimizer.param_groups
                ]
            else:
                frac = (float(self.last_epoch) - self.warmup_iters) / (
                    self.max_iters - self.warmup_iters
                )
                multiplier = max((1.0 - frac), 0.0) ** self.lr_pow
                return [
                    multiplier * group["initial_lr"]
                    for group in self.optimizer.param_groups
                ]



#
# Copyright 2021, Flyreel. All Rights Reserved.
# =============================================#
"""Omegaconf resolvers."""
from typing import Optional


class AttributeResolver:
    """An omegaconf resolver that does attribute lookups on an object."""

    def __call__(self, name: str):
        """Looks up the attribute with the given name."""
        return getattr(self, name)


class SESEMIConfigAttributes(AttributeResolver):
    """The attributes exposed to SESEMI configuration files.

    These attributes can be referenced in the config files by following the omegaconf syntax for
    custom resolvers. For example, ${sesemi:iterations_per_epoch} will reference the
    `iterations_per_epoch` attribute.

    Attributes:
        iterations_per_epoch: The number of training iterations per epoch if training data is
            available.
        max_iterations: The maximum number of training iterations if training data is available.
        num_gpus: The number of GPUs that will be used.
        num_nodes: The number of compute nodes that will be used.
    """

    iterations_per_epoch: Optional[int]
    max_iterations: Optional[int]
    num_gpus: int
    num_nodes: int



#
# Copyright 2021, Flyreel. All Rights Reserved.
# =============================================#
"""Omegaconf structured configurations."""
from dataclasses import dataclass, field
from typing import Any, Dict, List, Mapping, Optional
from omegaconf.dictconfig import DictConfig
from omegaconf.omegaconf import MISSING
from enum import Enum


class RunMode(Enum):
    """The mode to use when running the training CLI.

    Attributes:
        FIT: Performs training and validation.
        VALIDATE: Generates the validation metrics.
        TEST: Generates the test metrics.
    """

    FIT = "fit"
    VALIDATE = "validate"
    TEST = "test"


class DatasetConfig(DictConfig):
    """The SESEMI dataset configuration.

    Attributes:
        name: The name of the dataset in the registry.
        root: The absolute or relative (with respect to the run's data root) path to the dataset.
            If this is None, then the run's `data_root` attribute will be used instead.
        subset: Can either be a the string name of the subset, a list of subsets, or None (null)
            to indicate the full set.
        image_transform: An optional callable transform configuration that is applied to images.
    """

    name: str
    root: Optional[str]
    subset: Any
    image_transform: Any
    _target_: str

    def __init__(self, defaults: Optional[Mapping[str, Any]] = None):
        super().__init__(
            {
                "root": None,
                "subset": None,
                "image_transform": None,
                "_target_": "sesemi.dataset",
            }
        )
        if defaults is not None:
            self.update(defaults)


@dataclass
class DataLoaderConfig:
    """The data loader configuration.

    Most of the attributes are taken directly from PyTorch's DataLoader object.

    Attributes:
        dataset: The dataset configuration.
        batch_size: An optional batch size to use for a PyTorch data loader. Cannot be set with
            `batch_size_per_gpu`.
        batch_size_per_gpu: An optional batch size per GPU to use. Cannot be set with `batch_size`.
        shuffle: Whether to shuffle the dataset at each epoch.
        sampler: An optional sampler configuration.
        batch_sampler: An optional batch sampler configuration.
        num_workers: The number of workers to use for data loading (0 means use main process).
        collate_fn: An optional collation callable configuration.
        pin_memory: Whether to pin tensors into CUDA.
        drop_last: Whether to drop the last unevenly sized batch.
        timeout: The timeout to use get data batches from workers.
        worker_init_fn: An optional callable that is invoked for each worker on initialization.

    References:
        * https://pytorch.org/docs/1.6.0/data.html?highlight=dataloader#torch.utils.data.DataLoader
    """

    dataset: DatasetConfig = field(default_factory=DatasetConfig)
    batch_size: Optional[int] = None
    batch_size_per_gpu: Optional[int] = None
    shuffle: bool = False
    sampler: Optional[Any] = None
    batch_sampler: Optional[Any] = None
    num_workers: Optional[int] = 0
    collate_fn: Optional[Any] = None
    pin_memory: bool = False
    drop_last: Optional[bool] = False
    timeout: float = 0
    worker_init_fn: Optional[Any] = None
    _target_: str = "sesemi.DataLoader"


@dataclass
class DataConfig:
    """The data group configuration.

    Attributes:
        train: An optional dictionary of data loader configurations. This configuration is directly
            mapped into dictionaries of data batches.
        val: An optional data loader configuration to use during validation.
        test: An optional data loader configuration to use for testing.
    """

    train: Optional[Dict[str, DataLoaderConfig]] = None
    val: Optional[DataLoaderConfig] = None
    test: Optional[DataLoaderConfig] = None


@dataclass
class RunConfig:
    """The run configuration.

    Attributes:
        seed: An optional random seed used on initialization.
        num_epochs: The number of training epochs to train. Cannot be set with `num_iterations`.
        num_iterations: The number of training iterations to run. Cannot be set with `num_epochs`.
        gpus: Either an integer specifying the number of GPUs to use, a list of GPU
            integer IDs, a comma-separated list of GPU IDs, or None to train on the CPU. Setting
            this to -1 uses all GPUs and setting it to 0 also uses the CPU.
        num_nodes: The number of nodes to use during training (defaults to 1).
        accelerator: Supports either "dp" or "ddp" (the default).
        batch_size_per_gpu: An optional default batch size per GPU to use with all data loaders.
        data_root: The directory to use as the parent of relative dataset root directories
            (see `DatasetConfig`).
        id: The identifier to use for the run.
        dir: The directory to store run outputs (e.g. logs, configurations, etc.).
        mode: The run's mode.
        resume_from_checkpoint: An optional checkpoint path to restore trainer state.
        pretrained_checkpoint_path: An optional checkpoint path to load pretrained model weights.
    """

    seed: Optional[int] = None
    num_epochs: Optional[int] = None
    num_iterations: Optional[int] = None
    gpus: Any = -1
    num_nodes: int = 1
    accelerator: Optional[str] = None
    batch_size_per_gpu: Optional[int] = None
    data_root: str = "./data"
    id: str = "default"
    dir: str = "./runs"
    mode: RunMode = RunMode.FIT
    resume_from_checkpoint: Optional[str] = None
    pretrained_checkpoint_path: Optional[str] = None


@dataclass
class LearnerConfig:
    """A base learner configuration."""

    _target_: str = MISSING


@dataclass
class SESEMIBaseConfig:
    """The base SESEMI configuration.

    Attributes:
        run: The run config.
        data: The data config.
        learner: The learner config.
        trainer: Optional additional parameters that can be passed to a PyTorch Lightning Trainer
            object.

    References:
        * https://pytorch-lightning.readthedocs.io/en/latest/common/trainer.html#trainer-class-api
    """

    run: RunConfig = RunConfig()
    data: DataConfig = DataConfig()
    learner: LearnerConfig = LearnerConfig()
    trainer: Any = None


@dataclass
class LossHeadConfig:
    """The loss head configuration.

    Attributes:
        head: A loss head configuration that can be instantiated.
        scheduler: An optional learning rate scheduler that can be instantiated.
        reduction: The loss reduction method to use (e.g. "mean" or "sum").
        scale_factor: The loss scale factor.
    """

    head: Any
    scheduler: Any = None
    reduction: str = "mean"
    scale_factor: float = 1.0


@dataclass
class LossCallableConfig:
    """A callable loss configuration.

    Attributes:
        callable: An callable configuration that can be instantiated.
        scheduler: An optional learning rate scheduler that can be instantiated.
        reduction: The loss reduction method to use (e.g. "mean" or "sum").
        scale_factor: The loss scale factor.
    """

    callable: Any = MISSING
    scheduler: Any = None
    reduction: str = "mean"
    scale_factor: float = 1.0


@dataclass
class LRSchedulerConfig:
    """A learning rate scheduler configuration.

    Attributes:
        scheduler: A learning rate scheduler that can be instantiated.
        frequency: The frequency of learning rate updates as a multiple of `interval`.
        interval: Specifies the learning rate update interval. Can be either "step" or "epoch".
        monitor: An optional metric that is monitored for specific kinds of schedulers.
        strict: Wehether to enforce that the `monitor` metric should exist on updates.
        name: An optional name that can be used by certain learning rate callbacks.
    """

    scheduler: Any
    frequency: int = 1
    interval: str = "step"
    monitor: Optional[str] = None
    strict: bool = True
    name: Optional[str] = None


@dataclass
class EMAConfig:
    """A configuration for EMA models.

    Attributes:
        decay: The decay rate of the exponential moving average.
    """

    decay: float = 0.999


@dataclass
class ClassifierModelConfig:
    """The classifier learner's model configuration.

    Attributes:
        backbone: A backbone config that can be instantiated.
        supervised_loss: A callable loss config.
        regularization_loss_heads: An optional dictionary of loss head configs.
        ema: An optional config for the ema decay coefficient.
    """

    backbone: Any = MISSING
    supervised_loss: LossCallableConfig = LossCallableConfig()
    regularization_loss_heads: Optional[Dict[str, LossHeadConfig]] = None
    ema: Optional[EMAConfig] = None


@dataclass
class ClassifierHParams:
    """The classifier hyperparameters.

    Attributes:
        num_classes: The number of classes to use.
        model: The classifier's model config.
        optimizer: An optimizer config that can be instantiated.
        lr_scheduler: An optional learning rate scheduler config.
    """

    num_classes: int = MISSING
    model: ClassifierModelConfig = ClassifierModelConfig()
    optimizer: Any = MISSING
    lr_scheduler: Optional[LRSchedulerConfig] = None


@dataclass
class ClassifierConfig(LearnerConfig):
    """The classifier configuration.

    Attributes:
        hparams: The classifier's hyperparameters.
    """

    hparams: ClassifierHParams = ClassifierHParams()
    _target_: str = "sesemi.Classifier"


@dataclass
class SESEMIPseudoDatasetConfig:
    """The pseudo-dataset generation config.

    Attributes:
        checkpoint_path: The path to the checkpoint to load.
        seed: The random seed used on initialization.
        output_dir: The directory to save the pseudo-labeled dataset.
        dataset: The dataset configuration.
        preprocessing_transform: The preprocessing transform.
        test_time_augmentation: The test-time augmentation that takes an image
            and returns a list of augmented versions of that image.
        postaugmentation_transform: A transform to apply after test-time
            augmentations and which should return a tensor.
        image_getter: The function to extract the source image from
            an example in the dataset.
        gpus: Either an integer specifying the number of GPUs to use, a list of
            GPU integer IDs, a comma-separated list of GPU IDs, or None to
            train on the CPU. Setting this to -1 uses all GPUs and setting it
            to 0 also uses the CPU.
        batch_size: The data loading and inference batch size.
        num_workers: The number of workers to use for data loaders.
        symlink_images: Whether to use symlinks for images in the
            pseudo-labeled dataset rather than copying the image.
    """

    checkpoint_path: str = MISSING
    seed: int = 42
    output_dir: str = MISSING
    dataset: DatasetConfig = field(default_factory=DatasetConfig)
    preprocessing_transform: Any = None
    test_time_augmentation: Any = None
    postaugmentation_transform: Any = None
    image_getter: Any = None
    gpus: Any = -1
    batch_size: int = 16
    num_workers: int = 6
    symlink_images: bool = True


@dataclass
class SESEMIInferenceConfig:
    """The inference operation config.

    Attributes:
        checkpoint_path: The path to the saved checkpoint.
        no_cuda: Whether to disable CUDA.
        data_dir: Path to test dataset with one or more subdirs containing images.
        batch_size: Mini-batch size.
        workers: Number of data loading workers.
        oversample: Whether to oversample.
        ncrops: Number of crops to oversample.
        topk: Return topk predictions.
        resize: Resize smaller edge to this resolution while maintaining
            aspect ratio.
        crop_dim: Dimension for center or multi cropping.
        outfile: File to write predictions to.
    """

    checkpoint_path: str = MISSING
    no_cuda: bool = False
    data_dir: str = MISSING
    batch_size: int = 16
    workers: int = 6
    oversample: bool = False
    ncrops: int = 5
    topk: int = 1
    resize: int = 256
    crop_dim: int = 224
    outfile: str = MISSING



from os import path
from setuptools import find_packages, setup

with open("requirements.txt", "r") as f:
    requirements = f.read().strip().splitlines()

with open(
    path.join(path.abspath(path.dirname(__file__)), "README.md"), encoding="utf-8"
) as f:
    long_description = f.read()

setup(
    name="sesemi",
    version="0.3.0",
    description="Image Classification with Self-Supervised Regularization",
    author="Flyreel AI",
    author_email="ai@flyreel.co",
    packages=find_packages(include=["sesemi*"]),
    entry_points={"console_scripts": ["open_sesemi = sesemi.trainer.cli:open_sesemi"]},
    url="https://github.com/FlyreelAI/sesemi/",
    include_package_data=True,
    install_requires=requirements,
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Science/Research",
        "Development Status :: 4 - Beta",
    ],
)



#
# Copyright 2021, Flyreel. All Rights Reserved.
# =============================================#
"""Top-level module imports."""

# NOTE: Imports PyTorch Lightning first to mitigate an import error.
import pytorch_lightning as _pl

from torch.utils.data import DataLoader

from .datasets import dataset
from .models.backbones.timm import PyTorchImageModels
from .schedulers.lr import PolynomialLR
from .learners import Classifier

import torchvision.transforms as T



"""Specialized residual networks."""
from typing import Optional
from torch import nn

import torch.nn.functional as F

from .base import Backbone


class CIFARResidualBlock(nn.Module):
    """A residual block for the CIFARResNet backbone."""

    def __init__(
        self,
        in_channels: int,
        out_channels: Optional[int] = None,
        downsample: bool = False,
    ):
        """Initializes the residual block.

        Args:
            in_channels: The number of input channels.
            out_channels: The number of output channels. Defaults to the number of
                inputs.
            downsamples: Whether to downsample the inputs using a stride of 2.
        """
        super().__init__()
        if out_channels is None:
            out_channels = in_channels

        self.in_channels = in_channels
        self.out_channels = out_channels

        self.conv1 = nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=3,
            stride=2 if downsample else 1,
            padding=1,
            bias=False,
        )
        self.bn1 = nn.BatchNorm2d(out_channels)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = nn.Conv2d(
            in_channels=out_channels,
            out_channels=out_channels,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=False,
        )
        self.bn2 = nn.BatchNorm2d(out_channels)
        self.downsample = downsample

    def forward(self, inputs):
        conv1 = self.conv1(inputs)
        bn1 = self.bn1(conv1)
        relu1 = self.relu(bn1)
        conv2 = self.conv2(relu1)
        residual = self.bn2(conv2)

        if self.downsample:
            xi = F.interpolate(
                inputs, size=(residual.shape[2], residual.shape[3]), mode="nearest"
            )
        else:
            xi = inputs

        channel_difference = self.out_channels - self.in_channels
        if channel_difference > 0:
            xi = F.pad(
                xi,
                (0, 0, 0, 0, 0, channel_difference, 0, 0),
                mode="constant",
                value=0,
            )

        y = residual + xi
        outputs = self.relu(y)
        return outputs


class CIFARDeepResidualBlock(nn.Module):
    """A deep residual block for the CIFARResNet backbone."""

    def __init__(
        self,
        num_residual_blocks: int,
        in_channels: int,
        out_channels: Optional[int] = None,
        downsample: bool = False,
    ):
        """Initializes the deep residual block.

        Args:
            num_residual_blocks: The number of residual block to chain together.
            in_channels: The number of input channels.
            out_channels: The number of output channels. Defaults to the number of
                inputs.
            downsamples: Whether to downsample the inputs using a stride of 2.
        """
        super().__init__()
        if out_channels is None:
            out_channels = in_channels

        self.num_residual_blocks = num_residual_blocks
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.downsample = downsample

        block_layers = [
            CIFARResidualBlock(
                in_channels=in_channels,
                out_channels=out_channels,
                downsample=downsample,
            )
        ]
        for _ in range(num_residual_blocks - 1):
            block_layers.append(CIFARResidualBlock(in_channels=out_channels))

        self.block = nn.Sequential(*block_layers)

    def forward(self, inputs):
        return self.block(inputs)


class CIFARResNet(Backbone):
    """A specialized ResNet designed for CIFAR10 and described in the referenced paper.

    @article{He2016DeepRL,
        title={Deep Residual Learning for Image Recognition},
        author={Kaiming He and X. Zhang and Shaoqing Ren and Jian Sun},
        journal={2016 IEEE Conference on Computer Vision and Pattern Recognition (CVPR)},
        year={2016},
        pages={770-778}
    }

    See section 4.2 for the CIFAR-10 network design and analysis.
    """

    def __init__(
        self,
        linear_features: Optional[int] = None,
        n: int = 18,
        drop_rate: float = 0.0,
    ):
        """Initializes the ResNet.

        Args:
            linear_features: An optional specification for an additional linear layer
                to use at the end of the network. Otherwise the outputs are 64d and no
                linear layer is applied.
            n: This parameter defines a multiplier for the number of layers in use. The
                total layers not counting any added linear linear is 6 * n + 1. The
                default value for n would then result in 109 (6 * 18 + 1) layers. With
                an added linear layer this is 110.
            drop_rate: The dropout rate to use on the output features.
        """
        super().__init__()

        self.in_features = 3
        self.out_features = 64 if linear_features is None else linear_features

        self.conv1 = nn.Conv2d(
            in_channels=3,
            out_channels=16,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=False,
        )
        self.bn1 = nn.BatchNorm2d(16)
        self.relu = nn.ReLU(inplace=True)

        self.block1 = CIFARDeepResidualBlock(n, in_channels=16)
        self.block2 = CIFARDeepResidualBlock(
            n, in_channels=16, out_channels=32, downsample=True
        )
        self.block3 = CIFARDeepResidualBlock(
            n, in_channels=32, out_channels=64, downsample=True
        )

        self.pooling = nn.AdaptiveAvgPool2d(1)

        if linear_features is not None:
            self.linear_block: Optional[nn.Sequential] = nn.Sequential(
                nn.Linear(64, linear_features, bias=False),
                nn.BatchNorm1d(linear_features),
                nn.ReLU(inplace=True),
            )
        else:
            self.linear_block = None

        self.dropout = nn.Dropout(p=drop_rate, inplace=True)

    def forward(self, inputs):
        conv1 = self.conv1(inputs)
        bn1 = self.bn1(conv1)
        relu1 = self.relu(bn1)
        block1 = self.block1(relu1)
        block2 = self.block2(block1)
        block3 = self.block3(block2)
        features = self.pooling(block3)[..., 0, 0]

        if self.linear_block is not None:
            features = self.linear_block(features)

        outputs = self.dropout(features)
        return outputs



#
# Copyright 2021, Flyreel. All Rights Reserved.
# =============================================#
"""Utility functions."""
import numpy as np
import os, errno
import torch
import logging

from typing import Any, Dict, List, Optional, Union
from torch import Tensor
from torchvision.datasets import ImageFolder

from omegaconf.base import SCMode
from omegaconf.omegaconf import OmegaConf

from itertools import combinations
from torch import nn

from omegaconf import DictConfig

logger = logging.getLogger(__name__)


def reduce_tensor(tensor: Tensor, reduction: Optional[str] = None) -> Tensor:
    """Reduces a tensor using the given mode.

    Args:
        reduction: An optional method to use when reducing the tensor. Can be one of
            (mean, sum, none).
    """
    if reduction == "mean":
        return torch.mean(tensor)
    elif reduction == "sum":
        return torch.sum(tensor)
    elif reduction is None or reduction == "none":
        return tensor
    else:
        raise ValueError(f"unsupported reduction method {reduction}")


def compute_num_gpus(gpus: Union[int, str, List[int]]) -> int:
    """Computes the number of GPUs being used.

    Args:
        gpus: Either an integer specifying the number of GPUs to use, a list of GPU
            integer IDs, a comma-separated list of GPU IDs, or None to train on the CPU. Setting
            this to -1 uses all GPUs and setting it to 0 also uses the CPU.

    Returns:
        The number of GPUs that will be used.
    """
    num_gpus = 0
    if gpus is not None:
        num_available_gpus = torch.cuda.device_count()
        if isinstance(gpus, int):
            num_gpus = gpus if gpus >= 0 else num_available_gpus
        elif isinstance(gpus, str):
            gpu_ids = [x.strip() for x in gpus.split(",")]
            if len(gpu_ids) > 1:
                num_gpus = len(gpu_ids)
            elif len(gpu_ids) == 1:
                if int(gpu_ids[0]) < 0:
                    num_gpus = num_available_gpus
                else:
                    num_gpus = 1
        else:
            num_gpus = len(gpus)
    return num_gpus


def compute_device_names(gpus: Union[int, str, List[int]]) -> List[str]:
    """Computes the device names for the given GPU config.

    Args:
        gpus: Either an integer specifying the number of GPUs to use, a list of GPU
            integer IDs, a comma-separated list of GPU IDs, or None to train on the CPU. Setting
            this to -1 uses all GPUs and setting it to 0 also uses the CPU.

    Returns:
        The torch devices that will be used.
    """
    if gpus is not None:
        num_available_gpus = torch.cuda.device_count()
        if isinstance(gpus, int):
            num_gpus = gpus if gpus >= 0 else num_available_gpus
            return [f"cuda:{i}" for i in range(num_gpus)]
        elif isinstance(gpus, str):
            gpu_ids = [x.strip() for x in gpus.split(",")]
            if len(gpu_ids) > 1:
                return [f"cuda:{x}" for x in gpu_ids]
            elif len(gpu_ids) == 1:
                if int(gpu_ids[0]) < 0:
                    return [f"cuda:{i}" for i in range(num_available_gpus)]
                else:
                    return [f"cuda:{gpu_ids[0]}"]
        else:
            return [f"cuda:{x}" for x in gpus]
    return ["cpu"]


def sigmoid_rampup(curr_iter: int, rampup_iters: int) -> float:
    """Computes the sigmoid ramp-up value.

    Exponential rampup from https://arxiv.org/abs/1610.02242

    Args:
        curr_iter: The current iteration (step).
        rampup_iters: The number of iterations to ramp-up.

    Returns:
        An increasing value between 0 and 1.
    """
    if rampup_iters == 0:
        return 1.0
    else:
        current = np.clip(curr_iter, 0.0, rampup_iters)
        phase = 1.0 - current / rampup_iters
        return float(np.exp(-5.0 * phase * phase))


def assert_same_classes(datasets: List[ImageFolder]):
    """Checks that the image folder datasets have the same classes.

    Args:
        datasets: The datasets to ensure have the same classes.
    """
    if len(datasets) == 1:
        return True
    same_classes = [
        x.class_to_idx == y.class_to_idx for x, y in combinations(datasets, r=2)
    ]
    assert all(
        same_classes
    ), f"The following have mismatched subdirectory names. Check the `Root location`.\n{datasets}"


def validate_paths(paths: List[str]):
    """Validates that the paths exist.

    Args:
        paths: The list of filesystem paths.
    """
    for path in paths:
        if not os.path.exists(path):
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), path)


def load_checkpoint(model: nn.Module, checkpoint_path: str, strict: bool = False):
    """Loads the classifier checkpoint.

    Args:
        model: The classifier.
        checkpoint_path: The path to the classifier's checkpoint.
    """
    logger.info(f"Loading {checkpoint_path}")
    logger.info("")
    with open(checkpoint_path, "rb") as f:
        checkpoint = torch.load(f)

    pretrained_state_dict = checkpoint["state_dict"]
    pretrained_state_dict.pop("best_validation_top1_accuracy", None)

    if not strict:
        current_state_dict = model.state_dict()
        if "fc.weight" in pretrained_state_dict:
            if "fc.weight" not in current_state_dict or (
                pretrained_state_dict["fc.weight"].shape
                != current_state_dict["fc.weight"].shape
            ):
                pretrained_state_dict.pop("fc.weight")
                pretrained_state_dict.pop("fc.bias")

        incompatible_keys = model.load_state_dict(pretrained_state_dict, strict=False)
        if incompatible_keys.missing_keys:
            logger.info("missing keys:")
            logger.info("---")
            logger.info("\n".join(incompatible_keys.missing_keys))
            logger.info("")

        if incompatible_keys.unexpected_keys:
            logger.info("unexpected keys:")
            logger.info("---")
            logger.info("\n".join(incompatible_keys.unexpected_keys))
            logger.info("")
    else:
        model.load_state_dict(pretrained_state_dict, strict=True)


def copy_config(
    config: DictConfig, structured_config_mode: SCMode = SCMode.DICT
) -> Union[Dict[str, Any], DictConfig]:
    """Copies an omegaconf configuration by resolving interpolatons.

    Args:
        config: The config to copy.
        structured_config_mode: The format to return.

    Returns:
        The copied config.
    """
    return OmegaConf.to_container(
        config, resolve=True, structured_config_mode=structured_config_mode
    )
